from datetime import datetime as dt

from .base import BaseStateHandler



class SQLAlchemyState(BaseStateHandler):

	def __init__(self, uri) -> None:
		super().__init__()
		self.uri = uri
		self._validate_uri()
		self.__tables_created = False

		try: # make sure required packages are installed
			import sqlalchemy
			import sqlalchemy_utils
		except ImportError:
			msg = f"Please install sqlalchemy and sqlalchemy-utils to use '{self.__class__.__name__}' class"
			print("=="*25)
			print(msg)
			print("=="*25)
			raise ImportError(msg)


	def _validate_uri(self):
		from sqlalchemy.engine.url import make_url
		url_validation = make_url(self.uri)

		if url_validation.drivername is None:
			raise ValueError("Invalid sqlalchemy URL. driver name missing")

		if url_validation.drivername != 'sqlite': # sqlite does not have the below
			if url_validation.host is None:
				raise ValueError("Invalid sqlalchemy URL. host name missing")

			if url_validation.username is None or url_validation.password is None:
				raise ValueError("Invalid sqlalchemy URL. credentials missing")

		if url_validation.database is None:
			raise ValueError("Invalid sqlalchemy URL. database name missing")


	def _ensure_create_table(self):
		if self.__tables_created is True:
			return
		self.__tables_created = True

		from sqlalchemy import create_engine, MetaData
		from sqlalchemy import Table, Column, String, Text, DateTime, Boolean
		from sqlalchemy import select, insert, update
		from sqlalchemy_utils import database_exists, create_database

		self._engine = create_engine(self.uri)
		if not database_exists(self._engine.url):
			create_database(self._engine.url)
		self._meta = MetaData()

		# sqlalchemy table definitions
		self.fp_apps = Table(
			'fp_apps', self._meta,
			Column('app_id', String(50), primary_key=True),
			Column('app_unique_info', String(255)),
			Column('restart_dt', DateTime)
		)

		self.fp_state = Table(
			'fp_state', self._meta,
			Column('app_id', String(50), primary_key=True),
			Column('signature', String(50), primary_key=True),
			Column('readable', String(200)),
			Column('log', Text),
			Column('err', Text),
			Column('start_dt', DateTime),
			Column('end_dt', DateTime),
			Column('disabled', Boolean),
		)
		self._meta.create_all(self._engine)

		# update info to fp_apps table
		info_str = '\n'.join(self._cur_app_unique_info)
		restart_dt = dt.now()
		with self._engine.connect() as conn:
			app_row = conn.execute(select(self.fp_apps).where(self.fp_apps.c.app_id == self._cur_app_unique_info_hash)).all()
			if len(app_row) > 0:
				update(self.fp_apps).where(self.fp_apps.c.app_id == self._cur_app_unique_info_hash).values(restart_dt=restart_dt)
				if info_str != app_row[0].app_unique_info:
					print("=="*25)
					print(f"{self.__class__.__name__}: HASH COLLISION")
					print("=="*25)
			else:
				conn.execute(insert(self.fp_apps).values(app_id=self._cur_app_unique_info_hash, app_unique_info=info_str, restart_dt=restart_dt))
				conn.commit()


	def save_job_logs(self, job_obj):
		self._ensure_create_table()
		from sqlalchemy import select, insert, update
		signature = job_obj.signature_hash()
		logs = job_obj._logs_to_dict()

		with self._engine.connect() as conn:
			stmt = select(self.fp_state).where(self.fp_state.c.signature == signature, self.fp_state.c.app_id == self._cur_app_unique_info_hash)
			db_job = conn.execute(stmt).all()
			if len(db_job) == 1: # has to return only 1 element as we are querying on primary keys
				# print(db_job[0])
				update_stmt = update(self.fp_state).where(self.fp_state.c.signature == signature, self.fp_state.c.app_id == self._cur_app_unique_info_hash)

			else: # no state exists. insert!
				update_stmt = insert(self.fp_state)

			conn.execute(update_stmt.values(
				app_id=self._cur_app_unique_info_hash,
				signature=signature,
				readable=job_obj.func_signature(),
				log=logs.get('log'),
				err=logs.get('err'),
				start_dt=logs.get('start'),
				end_dt=logs.get('end'),
				disabled=job_obj.is_disabled
			))
			conn.commit()


	def restore_all_job_logs(self, jobs_list):
		self._ensure_create_table()
		from sqlalchemy import select, delete

		with self._engine.connect() as conn:
			db_states = conn.execute(select(self.fp_state).where(self.fp_state.c.app_id == self._cur_app_unique_info_hash)).all()

		states = {}
		for s in db_states:
			states[s.signature] = s

		found_states = []
		for j in jobs_list.copy(): # work on a shallow copy of this list - safer in case the list changes. TODO: maybe use locks instead?
			signature = j.signature_hash()
			if signature in states:
				st = states[signature]
				j._logs_from_dict({
					'log': st.log,
					'err': st.err,
					'start': st.start_dt,
					'end': st.end_dt,
				})
				if st.disabled:
					j.disable()
				found_states.append(signature)
				# print("restored", j)

		# clean up other states that did not match current jobs list (possibly stale)
		with self._engine.connect() as conn:
			for sig, st in states.items():
				if sig not in found_states:
					conn.execute(delete(self.fp_state).where(self.fp_state.c.signature == sig))
			conn.commit()

		print("* scheduler state restored from database *")



from flask import Flask
from flask_production import TaskScheduler
from flask_production.plugins import TaskMonitor

import time
import json
import pytest

MONITOR_NAME = "Web Test"

app = Flask(__name__)
sched = TaskScheduler()
monitor = TaskMonitor(app, sched=sched, display_name=MONITOR_NAME)

toggle = False

@app.route("/", methods=['GET'])
def main():
	return 'Main dummy page'


def wash_car():
	"""
	This is a dummy job that is scheduled to wash my car
	Note: objects in the mirror are closer than they appear
	"""
	global toggle
	toggle = not toggle
	if toggle:
		count = 50
		while count > 0:
			time.sleep(0.1)
			print("washing..\n")
			count -= 1
		print("The car was washed")
	else:
		time.sleep(1)
		raise Exception("car wash failed!")

def another_task():
	print("another_task")


@pytest.fixture
def client():
	with app.test_client() as c:
		yield c


def test_webservice(client):
	assert(client.get("/").status_code==200)

def test_blankpage(client):
	homepage = client.get("/{}".format(monitor._endpoint))
	assert(homepage.status_code==200)
	assert(homepage.data.decode().lower()=='nothing here')


def test_monitor_homepage(client):
	sched.every("day").at("8:00").do(another_task)
	sched.every(20).do_parallel(wash_car)
	sched.every(30).do_parallel(lambda: wash_car())
	# CherryFlask(app, sched).run() # unused

	homepage = client.get("/{}".format(monitor._endpoint))
	assert(homepage.status_code==200)
	html_text = homepage.data.decode(errors='ignore').lower()
	assert("lambda" in html_text)
	assert("wash_car" in html_text)
	assert("another_task" in html_text)


def test_monitor_jobpage(client):
	jobpage = client.get("/{}/0".format(monitor._endpoint))
	assert(jobpage.status_code==200)
	html_text = jobpage.data.decode(errors='ignore').lower()
	assert("another_task" in html_text)
	assert("lambda" not in html_text)
	assert("logs" in html_text)
	assert("next run in" in html_text)


def test_monitor_rerun_btn(client):
	sched.every(30).do_parallel(another_task)
	res = client.post("/{}/rerun".format(monitor._endpoint), json={'jobid':0}) # missing api_token causes call to be blocked
	assert(res.status_code==200)
	assert("blocked" in res.data.decode(errors='ignore').lower())
	time.sleep(1)
	res = client.post("/{}/rerun".format(monitor._endpoint), json={'jobid':0, 'api_token': monitor._api_protection_token})
	assert("success" in res.data.decode(errors='ignore').lower())


def test_monitor_disable_btn(client):
	sched.every(30).do_parallel(another_task)
	payload = {'jobid':0, 'api_token': monitor._api_protection_token, 'disable': True}
	res = client.post("/{}/enable_disable".format(monitor._endpoint), json=payload)
	assert(res.status_code==200)
	assert(sched.get_job_by_id(0).is_disabled==True)
	time.sleep(1)
	payload['disable'] = False
	res = client.post("/{}/enable_disable".format(monitor._endpoint), json=payload)
	assert("success" in res.data.decode(errors='ignore').lower())
	assert(sched.get_job_by_id(0).is_disabled==False)



def test_monitor_all_json(client):
	sched.every("day").at("8:00").do(another_task)

	resp = client.get("/{}/json/all".format(monitor._endpoint), content_type='application/json')
	respdict = json.loads(resp.data.decode('utf8'))
	assert('success' in respdict)
	assert(isinstance(respdict['success'], list))
	assert(isinstance(respdict['success'][0], dict))
	assert(isinstance(respdict['success'][0]['logs'], dict))


def test_monitor_one_json(client):
	sched.every("day").at("8:00").do(another_task)

	resp = client.get("/{}/json/0".format(monitor._endpoint), content_type='application/json')
	respdict = json.loads(resp.data.decode('utf8'))
	assert('success' in respdict)
	assert(isinstance(respdict['success'], dict))
	assert(respdict['success']['jobid']==0)


def test_monitor_summary(client):
	sched.every("day").at("8:00").do(another_task)

	all_resp = client.get("/{}/json/all".format(monitor._endpoint), content_type='application/json')
	all_respdict = json.loads(all_resp.data.decode('utf8'))

	resp = client.get("/{}/json/summary".format(monitor._endpoint), content_type='application/json')
	respdict = json.loads(resp.data.decode('utf8'))
	assert('success' in respdict)
	assert(respdict['success']['name']==MONITOR_NAME)
	assert(respdict['success']['summary']['count']==len(respdict['success']['details']))
	assert(respdict['success']['summary']['errors']==0)

	assert(len(all_respdict['success'])==len(respdict['success']['details']))



import os, shutil
import time
import threading

from flask import Flask
import requests
from flask_production import CherryFlask, TaskScheduler, TaskMonitor
from flask_production.state import FileSystemState

CUR_APP_DATA_DIR_PATH = FileSystemState()._get_current_app_data_directory()

MONITOR_NAME = "Cherrypy Test"
TEST_PORT = 5555


def task():
	time.sleep(1)
	print("Done")


app = Flask(__name__)
sched = TaskScheduler()
monitor = TaskMonitor(app, sched=sched, display_name=MONITOR_NAME)

cherry = CherryFlask(app, sched)

thrd = threading.Thread(target=lambda: cherry.run(port=TEST_PORT))


@app.route("/", methods=['GET'])
def main():
	return 'Main dummy page'


def setup_module(module):
	thrd.start()
	time.sleep(1)


def teardown_module(module):
	cherry.stop()
	thrd.join()
	time.sleep(1)
	if os.path.isdir(CUR_APP_DATA_DIR_PATH):
		shutil.rmtree(CUR_APP_DATA_DIR_PATH)


def test_cherry_simple():
	res = requests.get(f"http://localhost:{TEST_PORT}/")
	assert(res.status_code==200)


def test_cherry_sched():
	res = requests.get(f"http://localhost:{TEST_PORT}/{monitor._endpoint}/json/summary")
	assert(res.status_code==200)
	assert('error' in res.json())

	sched.every(20).do(task)

	res = requests.get(f"http://localhost:{TEST_PORT}/{monitor._endpoint}/json/summary")
	assert(res.status_code==200)
	res_json = res.json()
	assert('success' in res_json)
	assert(res_json['success']['summary']['count']==1)
	assert(len(res_json['success']['details'])==1)



import time
from datetime import timedelta, datetime as dt
from monthdelta import monthdelta
from dateutil import tz
import re
import threading
import inspect
import hashlib
import traceback

from . import print_logger



class BadScheduleError(Exception):
	pass



def _get_eom(d):
	return ((d + monthdelta(1)).replace(day=1) - timedelta(days=1))

def _is_eom(d, hols):
	eom = _get_eom(d)
	return d.date() == eom.date()

def _is_eom_weekday(d, hols):
	eom = _get_eom(d)
	while eom.isoweekday() >= 6:
		eom -= timedelta(days=1)
	return d.date() == eom.date()

def _is_eom_businessday(d, hols):
	eom = _get_eom(d)
	while eom in hols or eom.isoweekday() >= 6:
		eom -= timedelta(days=1)
	return d.date() == eom.date()



RUNABLE_DAYS = {
	'day': lambda d, hols : True,
	'weekday': lambda d, hols : d.isoweekday() < 6,
	'weekend': lambda d, hols : d.isoweekday() > 5,
	'businessday': lambda d, hols : d not in hols and d.isoweekday() < 6,
	'holiday': lambda d, hols : d in hols or d.isoweekday() > 5,
	'trading-holiday': lambda d, hols : d in hols and d.isoweekday() < 6, # trading-holidays don't count if they fall on weekends
	# days of the week
	'monday': lambda d, hols: d.isoweekday() == 1,
	'tuesday': lambda d, hols: d.isoweekday() == 2,
	'wednesday': lambda d, hols: d.isoweekday() == 3,
	'thursday': lambda d, hols: d.isoweekday() == 4,
	'friday': lambda d, hols: d.isoweekday() == 5,
	'saturday': lambda d, hols: d.isoweekday() == 6,
	'sunday': lambda d, hols: d.isoweekday() == 7,
	# end of month
	'eom': _is_eom,
	'eom-weekday': _is_eom_weekday,
	'eom-businessday': _is_eom_businessday,
}




class Job(object):
	'''standard job class'''

	@classmethod
	def is_valid_interval(cls, interval, time_string):
		'''The generic Job class only supports these interval. See subclasses for others'''
		return interval in RUNABLE_DAYS and time_string is not None

	def __init__(self, jobid, every, at, func, kwargs):
		if str(every) == 'holiday':
			print("!!", "="*20, "!!")
			print("'holiday' interval  is deprecated and will be removed. \r\nUse 'weekend' and 'trading-holiday' instead")
			print("!!", "="*20, "!!")
		self.jobid = jobid
		self.interval = every
		self.time_string = at
		self.tzname = None
		self.func = func
		self.kwargs = kwargs
		self.is_running = False
		self.next_timestamp = 0
		self._is_disabled = False
		self._run_silently = False
		self._generic_err_handler = None
		self._err_handler = None
		self._func_src_code = inspect.getsource(self.func)
		# signatures for setters and getters
		self._func_signature = None
		self._job_signature_hash = None
		self._on_complete_cbs = []
		self._on_enable_cbs = []
		self._on_disable_cbs = []

	def init(self, calendar, tzname=None, generic_err_handler=None, startup_grace_mins=0):
		'''initialize extra attributes of job'''
		self.calendar = calendar
		self.tzname = tzname
		self._generic_err_handler = generic_err_handler
		self._startup_grace_mins = startup_grace_mins # look back on tasks if task scheduler just started
		self._run_info = print_logger._PrintLogger(tzname=tzname)
		self.schedule_next_run()
		return self

	def silently(self, run_silently=True):
		self._run_silently = run_silently
		return self

	def catch(self, err_handler):
		'''register job specific error handler'''
		self._err_handler = err_handler
		return self

	def register_callback(self, cb, cb_type="oncomplete"):
		'''
		register a callback function to be called when job completes
		- this callback function should expect a job object as argument
		'''
		if callable(cb) and len(inspect.signature(cb).parameters)==1:
			if cb_type == 'oncomplete':
				self._on_complete_cbs.append(cb)
			elif cb_type == 'ondisable':
				self._on_enable_cbs.append(cb)
			elif cb_type == 'onenable':
				self._on_disable_cbs.append(cb)
			else:
				raise ValueError("unsupported cb_type")
		return self

	@property
	def is_disabled(self):
		return self._is_disabled

	@is_disabled.setter
	def is_disabled(self, disable: bool):
		if disable is True:
			self._is_disabled = True
			self.next_timestamp = 0 # remove next schedule
			# call any registered ondisable callbacks
			for cb in self._on_disable_cbs:
				try:
					cb(self)
				except Exception as e:
					print("on-disable-cb-error:", str(e))
		else:
			self.schedule_next_run() # before enabling, we need to reschedule the job
			self._is_disabled = False
			# call any registered onenable callbacks
			for cb in self._on_enable_cbs:
				try:
					cb(self)
				except Exception as e:
					print("on-enable-cb-error:", str(e))

	def disable(self):
		self.is_disabled = True # calls setter
		return self

	def enable(self):
		self.is_disabled = False # calls setter
		return self

	# important datetime and timezone management methods
	def to_timestamp(self, d: dt):
		return d.timestamp()

	def to_datetime(self, t: float):
		return dt.fromtimestamp(t, tz=tz.gettz(self.tzname))

	def tz_now(self):
		return dt.now(tz=tz.gettz(self.tzname))

	def tz_dt(self, year, month, day, hour=0, minute=0, second=0, microsecond=0):
		d = dt(int(year), int(month), int(day), int(hour), int(minute), int(second), int(microsecond), tzinfo=tz.gettz(self.tzname))
		return tz.resolve_imaginary(d) # handles time that falls in the transition to/from daylight savings
	#

	def attach_upcoming_run_time(self, d: dt, just_ran: bool=False):
		'''
		attaches time to a datetime object based on the value of self.time_string
		note: self.time_string is usually a string of the form %H:%M, but can also be a list of time strings
		- returns earliest datetime that occurs in the future
		- returns None if no datetimes exist in the future
		'''
		dt_list = []
		def _add_dt(at):
			if not isinstance(at, str):
				raise BadScheduleError(f"Invalid time string '{self.time_string}'")
			h, m = at.split(':')
			dt_list.append(self.tz_dt(d.year, d.month, d.day, int(h), int(m)))

		if isinstance(self.time_string, (list,set,tuple)):
			for at in self.time_string:
				_add_dt(at)
		else:
			_add_dt(self.time_string)
		# to find an upcoming time, we need to filter dt_list based on current date and time
		now = self.tz_now().replace(second=0, microsecond=0)
		if just_ran:
			upcoming_list = [n for n in dt_list if n > now]
		else:
			now -= timedelta(minutes=self._startup_grace_mins)
			upcoming_list = [n for n in dt_list if n >= now]
		if len(upcoming_list) == 0:
			return None
		return min(upcoming_list)

	def schedule_next_run(self, just_ran=False):
		'''compute timestamp of the next run'''
		d = self.tz_now()
		upcoming = self.attach_upcoming_run_time(d, just_ran=just_ran)
		if not self._job_must_run_today() or upcoming is None:
			next_day = d + timedelta(days=1)
			while not self._job_must_run_today(next_day):
				next_day += timedelta(days=1)
			upcoming = self.attach_upcoming_run_time(next_day)

		self.next_timestamp = self.to_timestamp(upcoming)

	def _job_must_run_today(self, date=None):
		return RUNABLE_DAYS[self.interval](date or self.tz_now(), self.calendar)

	def is_due(self):
		'''test if job should run now'''
		return self.next_timestamp > 0 and (time.time() >= self.next_timestamp) and not self.is_running and not self.is_disabled

	def did_fail(self):
		'''test if job failed'''
		return self._run_info.error != ''

	def func_signature(self):
		'''create human readable function signature'''
		if self._func_signature is None:
			def readable_trim(s):
				if isinstance(s, (list,tuple)):
					return "[..]"
				elif isinstance(s, set):
					return "(..)"
				elif isinstance(s, dict):
					return "{..}"
				else:
					s_str = str(s)[:6] + ".." if len(str(s))>6 else str(s)
					return s_str.replace("<", "*").replace(">", "*") # escaping html
			arguments = ''
			if self.kwargs:
				arguments = '({})'.format(','.join(['{}={}'.format(k, readable_trim(v)) for k,v in self.kwargs.items()]))
			if self.func.__module__ == "__main__":
				self._func_signature = '{}{}'.format(self.func.__qualname__, arguments)
			else:
				self._func_signature = '{}.{}{}'.format(self.func.__module__, self.func.__qualname__, arguments)
		return self._func_signature

	def signature_hash(self):
		'''create unique job signature hash'''
		if self._job_signature_hash is None:
			sig = "{}-{}-{}.{}({})".format(
				self.interval,
				self.time_string,
				self.func.__module__,
				self.func.__qualname__,
				list(self.kwargs.values())
			)
			self._job_signature_hash = hashlib.sha1(sig.encode()).hexdigest()
		return self._job_signature_hash


	def _run(self, is_rerun: bool, kwargs: dict=None):
		'''this is an internal runner. see self.run() for more'''
		self.is_running = True
		try:
			if not self._run_silently: # add print statements
				print("========== [{:03}] - Job {} [{}] =========".format(
					self.jobid,
					"Rerun Start" if is_rerun else "Start",
					self.tz_now().strftime("%Y-%m-%d %H:%M:%S %Z")
				))
				print("Executing {}".format(self))
				print("*") # job log seperator
			start_time = time.time()

			kw = self.kwargs.copy() # start with default kwargs
			if isinstance(kwargs, dict):
				# if kwargs provided, update
				# however, job will fail if a key is provided that is not expected by the job
				# TODO: should we remove bad arguments here, or let it fail?
				kw.update(kwargs)

			return self.func(**kw)

		except Exception:
			traceback.print_exc()
			print("Job", self.func_signature(), "failed!")
			err_msg = "Error in {}\n\n\n{}".format(self.func_signature(), traceback.format_exc())
			self._run_info.set_error()
			try:
				if self._err_handler is not None:
					self._err_handler(err_msg) # job specific error callback registered through .catch()
				elif self._generic_err_handler is not None:
					self._generic_err_handler(err_msg) # generic error callback from scheduler
			except:
				traceback.print_exc()
		finally:
			# if the job was forced to rerun, we should not schedule the next run
			if not is_rerun:
				self.schedule_next_run(just_ran=True)
			if not self._run_silently: # add print statements
				print("*") # job log seperator
				print( "Finished in {:.2f} minutes".format((time.time()-start_time)/60))
				print(self)
				print("========== [{:03}] - Job {} [{}] =========".format(
					self.jobid,
					"Rerun End" if is_rerun else "End",
					self.tz_now().strftime("%Y-%m-%d %H:%M:%S %Z")
				))
			self.is_running = False


	def run(self, is_rerun: bool=False, kwargs: dict=None):
		'''
		begin job run
		- redirected all print statements to _PrintLogger
		- call error handlers if provided
		- execute registered callback functions
		'''
		with self._run_info.start_capture(): # captures all writes to stdout
			self._run(is_rerun=is_rerun, kwargs=kwargs)

		for cb in self._on_complete_cbs: # call any registered on-complete callbacks
			try:
				cb(self)
			except Exception as e:
				print("on-complete-cb-error:", str(e))


	def _next_run_dt(self):
		return self.to_datetime(self.next_timestamp) if self.next_timestamp!=0 else None

	def _logs_to_dict(self):
		return self._run_info.to_dict() if hasattr(self, '_run_info') else {}

	def _logs_from_dict(self, logs_dict):
		if hasattr(self, '_run_info'):
			self._run_info.from_dict(logs_dict)

	def to_dict(self):
		'''property to access job info dict'''
		return dict(
			jobid=self.jobid,
			func=self.func.__qualname__,
			signature=self.func_signature(),
			src=self._func_src_code,
			doc=self.func.__doc__,
			type=self.__class__.__name__,
			every=self.interval,
			at=self.time_string,
			tzname=self.tzname,
			is_running=self.is_running,
			is_disabled=self.is_disabled,
			next_run=self._next_run_dt(),
			logs=self._logs_to_dict(),
		)

	def __repr__(self):
		d = self._next_run_dt()
		return "{:10} [{:03}] | Next run = {} | {}".format(
			self.__class__.__name__,
			self.jobid,
			d.strftime("%Y-%m-%d %H:%M:%S %Z") if isinstance(d, dt) else 'Never',
			self.func_signature()
		)


class OneTimeJob(Job):
	'''type of job that runs only once'''

	@classmethod
	def is_valid_interval(cls, interval, time_string):
		try:
			dt.strptime(interval, "%Y-%m-%d")
			return True
		except:
			return False

	def schedule_next_run(self, just_ran=False):
		Y, m, d = self.interval.split('-')
		n = self.tz_dt(int(Y), int(m), int(d))
		upcoming = self.attach_upcoming_run_time(n, just_ran=just_ran)

		if upcoming is None:
			self.next_timestamp = 0
		else:
			self.next_timestamp = self.to_timestamp(upcoming)

	def is_due(self):
		if self.next_timestamp==0:
			return False
		return super().is_due()


class RepeatJob(Job):
	'''type of job that runs every n seconds'''

	@classmethod
	def is_valid_interval(cls, interval, time_string):
		return isinstance(interval, (int, float)) and time_string is None

	def schedule_next_run(self, just_ran=False):
		if not isinstance(self.interval, (int, float)) or self.interval <= 0:
			raise BadScheduleError("Illegal interval for repeating job. Expected number of seconds")

		if just_ran:
			self.next_timestamp += self.interval
		else:
			self.next_timestamp = time.time() + self.interval


class MonthlyJob(Job):
	'''
	type of job that can be scheduled to run once per month
	example interval 1st, 22nd, 30th
	limitation: we cannot intuitively handle dates >= 29 for all months
		- ex: 31st will fail for months having less than 31 days, 29th will fail for non leap-Feb
		- use 'self._strict_date' when handing dates >= 29:
			if self._strict_date == True:
				job is scheduled only on months which have the date (ex: 31st)
			elif self._strict_date == False:
				run on the last day of the month if date exceeds current month
	'''

	PATTERN = re.compile(r"^(\d{1,2})(st|nd|rd|th)$", re.IGNORECASE)

	def __init__(self, jobid, every, at, func, kwargs, strict_date):
		if not isinstance(strict_date, bool):
			raise BadScheduleError("call to .strict_date() required for monthly schedule. ex: .every('31st').strict_date(True)..")
		self._strict_date = strict_date
		super().__init__(jobid, every, at, func, kwargs)

	@classmethod
	def is_valid_interval(cls, interval, time_string):
		'''example intervals - 1st, 22nd, 30th'''
		match = cls.PATTERN.match(str(interval))
		return (match is not None and int(match.groups()[0]) <= 31) and time_string is not None

	def schedule_next_run(self, just_ran=False):
		interval = int(self.PATTERN.match(self.interval).groups()[0])
		sched_day = self.tz_now()
		upcoming = self.attach_upcoming_run_time(sched_day, just_ran=just_ran)
		# switch to next month if
		# - task just ran, or
		# - day has already passed, or
		# - day is today, but time has already passed
		# - day is after today, and today is end of month and time has already passed (ex: 31st while current month has 28 or 30 days)
		_pure_time_passed = upcoming is None
		day_passed = interval < sched_day.day # True if day already passed this month
		time_passed = interval == sched_day.day and _pure_time_passed
		last_day_case = interval > sched_day.day and _get_eom(sched_day).day == sched_day.day and _pure_time_passed

		if day_passed or time_passed or last_day_case:
			sched_day += monthdelta(1) # switch to next month

		# handle cases where the interval day doesn't occur in all months (ex: 31st)
		if interval > _get_eom(sched_day).day:
			if self._strict_date==False:
				interval = _get_eom(sched_day).day # if strict is false, run on what ever is last day of the month
			else: # strict
				while interval > _get_eom(sched_day).day: # run only on months which have the date
					sched_day += monthdelta(1)

		n = self.attach_upcoming_run_time(sched_day.replace(day=interval))
		self.next_timestamp = self.to_timestamp(n)

	def __repr__(self):
		r = super().__repr__()
		if self._strict_date:
			r = r.replace(self.__class__.__name__, self.__class__.__name__+"[strict]")
		return r


class AsyncJobWrapper(object):
	'''wrapper to run the job on a parallel thread'''

	def __init__(self, job):
		self.job = job
		self.proc = None

	def __repr__(self):
		r = self.job.__repr__()
		r = r.replace(self.job.__class__.__name__, self.job.__class__.__name__+"(*)") # (*) indicates asynchronous job
		return r

	def __getattr__(self, name):
		return self.job.__getattribute__(name)

	def is_due(self):
		return self.job.is_due()

	def run(self, *args, **kwargs):
		self.proc = threading.Thread(target=self.job.run, args=args, kwargs=kwargs)
		self.proc.daemon = True
		self.proc.start()



class NeverJob(Job):
	'''type of job that runs only on demand (using TaskMonitor plugin)'''

	@classmethod
	def is_valid_interval(cls, interval, time_string):
		return interval in ('on-demand', 'never') and time_string is None

	def schedule_next_run(self, just_ran=False):
		self.next_timestamp = 0

	def is_due(self):
		return False



import socket
import requests
# import json
import psutil
from flask import send_file

from .html_templates import * # pylint: disable=unused-wildcard-import




class ControlPanel:
	'''Automatically scan ports and consolidate many TaskMonitors'''

	def __init__(self,
		app,
		ports=[],
		external_addrs=[],
		page_refresh=60):

		self.app = app
		self.machine = socket.gethostname()
		self.local_ip = socket.gethostbyname(self.machine)
		self.ports = set(ports)
		self.external_addrs = set(external_addrs)
		self.page_refresh = page_refresh
		self.app.add_url_rule("/", view_func=self._render_monitors, methods=['GET'])
		self.app.add_url_rule("/static/<type>/<filename>", view_func=self.__serve_file, methods=['GET'])


	def __serve_file(self, type, filename):
		return send_file(os.path.join(WEB_FOLDER, type, filename))


	def scan(self, min_port=1000, max_port=10000, timeout=5):
		for conn in psutil.net_connections():
			if conn.status == "LISTEN" and conn.laddr.port >= min_port and conn.laddr.port <= max_port:
				m = self._get_taskmonitor(self.local_ip, conn.laddr.port, timeout=timeout)
				if m is not None:
					self.ports.add(conn.laddr.port)
				print('>> scanned', conn.laddr.port, "- found" if m is not None else "")


	def _get_taskmonitor(self, host, port, timeout=5):
		try:
			if port == 80:
				monitor_url = f"http://{host}/@taskmonitor" # need to add option to change this endpoint since task monitor has that option
			elif port == 443:
				monitor_url = f"https://{host}/@taskmonitor" # need to add option to change this endpoint since task monitor has that option
			else:
				monitor_url = f"http://{host}:{port}/@taskmonitor" # need to add option to change this endpoint since task monitor has that option

			res = requests.get(f"{monitor_url}/json/summary", timeout=timeout).json()
			# print(json.dumps(res, indent=4))
			res['port'] = port
			res['url'] = monitor_url
			return res
		except Exception:
			# print(e)
			return None


	def _iter_monitors(self):
		for port in self.ports:
			monitor = self._get_taskmonitor(self.local_ip, port)
			if monitor is not None:
				yield monitor
		for host, port in self.external_addrs:
			monitor = self._get_taskmonitor(host, port)
			if monitor is not None:
				yield monitor


	def _render_monitors(self):
		content = []
		tot_jobs = 0
		for monitor in self._iter_monitors():
			css = ['monitor-block']
			attrs = {}
			elem = ""
			if 'error' in monitor:
				elem = H(5, monitor['error']) + SPAN(str(monitor['url']))
				css.append('error-border')
				css.append('no-page')
				attrs['title'] = monitor['error']
			else:
				mon = monitor['success']
				# for backward compatibility, apply defaults
				if 'errors' not in mon['summary']:
					mon['summary']['errors'] = 0
				if 'running' not in mon['summary']:
					mon['summary']['running'] = 0

				err_msg_css = []
				running_msg_css = []
				if mon['summary']['errors'] > 0:
					css.append('error-border')
					err_msg_css = ['red']
				if mon['summary']['running'] > 0:
					running_msg_css = ['yellow']

				task_msg = f"tasks: {DIV(mon['summary']['count'])}"
				running_msg = f"running: {DIV(mon['summary']['running'], css=running_msg_css)}"
				error_msg = f"errors: {DIV(mon['summary']['errors'], css=err_msg_css)}"

				msg = f"{task_msg}  {running_msg}  {error_msg}"
				elem = SPAN(B(mon['name']), css=['block-title']) + SPAN(msg, css=['block-msg'])
				attrs['data-url'] = monitor['url']
				attrs['title'] = f"{mon['name']}\n{monitor['url']}"
			content.append(DIV(elem, css=css, attrs=attrs))
			tot_jobs += mon['summary']['count']

		wrapper = DIV(''.join(content), css='wrapper')
		header_txt = f"Control Panel"
		header = DIV(H(2, header_txt), css=['header-bar'])
		summary_txt = SMALL(f"Monitoring {tot_jobs} jobs")
		rerun_txt = SMALL(f"Auto-refresh in {SPAN(self.page_refresh, attrs={'id': 'refresh-msg'})} seconds")

		return HTML(''.join([
			STYLE_LINK('/static/css/dark_theme.css'),
			STYLE_LINK('/static/css/ctrl_panel.css'),
			header,
			summary_txt,
			rerun_txt,
			wrapper,
			SCRIPT(f'''let COUNT_DOWN = {self.page_refresh}'''),
			SCRIPT_SRC('/static/js/ctrl_panel.js')
		]), title=header_txt)



import os, glob, shutil
import time
import threading
import json
from datetime import datetime as dt, timedelta
from monthdelta import monthdelta
from dateutil.parser import parse as date_parse
from dateutil import tz
import pytest

from flask_production import TaskScheduler
from flask_production.jobs import Job
from flask_production.hols import TradingHolidays
from flask_production.sched import LOGGER, BadScheduleError
from flask_production.state import FileSystemState, SQLAlchemyState

CUR_APP_DATA_DIR_PATH = FileSystemState()._get_current_app_data_directory()


LOGGING_TEST_FILE = 'testlog.log'
DB_STATE_TEST_FILE = 'teststate.db'


def job(x, y):
	time.sleep(0.1)
	print(x, y)

def pretty_print(d):
	print(json.dumps(d, indent=4, default=str))


def teardown_function(function):
	for h in LOGGER.handlers:
		h.close()
		LOGGER.removeHandler(h)

	for f in glob.glob(LOGGING_TEST_FILE+'*'):
		if os.path.isfile(f):
			os.remove(f)


def teardown_module(module):
	time.sleep(1)
	if os.path.isdir(CUR_APP_DATA_DIR_PATH):
		shutil.rmtree(CUR_APP_DATA_DIR_PATH)



def test_registry():
	s = TaskScheduler()
	s.every("businessday").at("10:00").do(job, x="hello", y="world") # Job
	s.on('2019-05-16').do(job, x="hello", y="world") # OneTimeJob
	s.every(5).at("10:00").do(job, x="hello", y="world") # RepeatJob
	s.every('2nd').strict_date(False).at("10:00").do(job, x="hello", y="world") # MonthlyJob
	assert(len(s.jobs) == 4)


def test_badinterval():
	s = TaskScheduler()
	with pytest.raises(BadScheduleError):
		s.every("day").at(5).do(job, x="hello", y="world")
	with pytest.raises(BadScheduleError):
		s.every("day").at([-1,2]).do(job, x="hello", y="world")
	with pytest.raises(BadScheduleError):
		s.every("potato").at("10:00").do(job, x="hello", y="world")
	with pytest.raises(BadScheduleError):
		s.every("2020-02-30").at("10:00").do(job, x="hello", y="world") # OneTimeJob
	with pytest.raises(BadScheduleError):
		s.every(0).at("10:00").do(job, x="hello", y="world") # RepeatJob
	with pytest.raises(BadScheduleError):
		# error because .strict_date() is not called
		s.every("31st").at("10:00").do(job, x="hello", y="world") # MonthlyJob
	with pytest.raises(BadScheduleError):
		s.every("32nd").strict_date(False).at("10:00").do(job, x="hello", y="world") # MonthlyJob


def test_regular():
	d = dt.now().replace(hour=23, minute=59, second=0, microsecond=0)
	s = TaskScheduler()
	s.every("day").at("23:59").do(job, x="hello", y="world")
	assert (s.jobs[0].next_timestamp==dt.timestamp(d))


def test_day_of_week():
	now = dt.now()
	today_str = now.strftime("%A").lower() # day of the week
	in2sec_str = now.strftime("%H:%M")
	s = TaskScheduler()
	s.every(today_str).at(in2sec_str).do(job, x="hello", y=today_str)
	assert(len(s.jobs) == 1)
	time.sleep(0.5)
	s.check()
	# test if next run greater than 6 days, less than 8 days from now
	test_timestamp = time.time()
	assert s.jobs[0].next_timestamp > test_timestamp+(6*24*60*60)
	assert s.jobs[0].next_timestamp < test_timestamp+(8*24*60*60)


def test_holidays():
	s = TaskScheduler() # default holidays calendar
	s.every("businessday").at("10:00").do(job, x="hello", y="world")
	assert(s.jobs[0]._job_must_run_today(date_parse("2020-04-09"))==True)
	assert(s.jobs[0]._job_must_run_today(date_parse("2024-11-11"))==False) # Veterans Day is a US holiday

	s = TaskScheduler(holidays_calendar=TradingHolidays())
	s.every("businessday").at("10:00").do(job, x="hello", y="world")
	assert(s.jobs[0]._job_must_run_today(date_parse("2020-04-09"))==True)
	assert(s.jobs[0]._job_must_run_today(date_parse("2020-04-11"))==False) # saturday
	assert(s.jobs[0]._job_must_run_today(date_parse("2024-11-11"))==True) # test Custom financial calendar with Veterans Day removed

	s.every("trading-holiday").at("10:00").do(job, x="hello", y="world")
	assert(s.jobs[1]._job_must_run_today(date_parse("2020-01-01"))==True)
	assert(s.jobs[1]._job_must_run_today(date_parse("2020-01-02"))==False)


def test_multi_intraday(): # test list of timestamps for .at()
	sched = TaskScheduler()
	n1 = dt.now().replace(second=0, microsecond=0)
	n2 = (n1+timedelta(minutes=1))
	sched.every("day").at([n1.strftime("%H:%M"), n2.strftime("%H:%M")]).do(job, x="multi", y="intraday")
	assert(sched.jobs[0].next_timestamp==dt.timestamp(n1))
	sched.check()
	assert(sched.jobs[0].next_timestamp==dt.timestamp(n2))


def test_onetime():
	yesterday = dt.now() - timedelta(days=1)
	tomorrow = (dt.now() + timedelta(days=1)).replace(hour=23, minute=59, second=0, microsecond=0)
	sched = TaskScheduler()
	dead_job = sched.on(yesterday.strftime("%Y-%m-%d")).at("23:59").do(job, x="one", y="time")
	assert(dead_job.next_timestamp == 0)
	alive_job = sched.on(tomorrow.strftime("%Y-%m-%d")).at("23:59").do(job, x="one", y="time")
	assert (alive_job.next_timestamp==dt.timestamp(tomorrow))
	assert(len(sched.jobs) == 2)


def test_never():
	sched = TaskScheduler()
	sched.every('on-demand').do(job, x="never", y="never")
	sched.on('never').do(job, x="never", y="never")
	for j in sched.jobs:
		assert (j.next_timestamp==0)
	assert(len(sched.jobs) == 2)
	sched.check()
	assert(len(sched.jobs) == 2)
	sched.jobs[0].run() # run the job directly
	sched.rerun(sched.jobs[1].jobid) # rerun -> runs as a thread
	time.sleep(0.1)
	assert(len(sched.jobs) == 2)
	for j in sched.jobs: # assert again that they were not rescheduled after running
		assert (j.next_timestamp==0)


def test_eom():
	s = TaskScheduler()
	s.every("eom").do(job, x="hello", y="eom")
	eom = dt.fromtimestamp(s.jobs[0].next_timestamp)
	assert((eom + timedelta(days=1)).day == 1)

	s.every("eom-weekday").do(job, x="hello", y="eom")
	eom = dt.fromtimestamp(s.jobs[1].next_timestamp)
	assert(eom.isoweekday() < 6) # make sure it's a weekday
	eom += timedelta(days=1) # count up to the first if next month
	while eom.day != 1:
		assert(eom.isoweekday() >= 6) # if there are any days between eom-weekday and 1st of next month, those have to be weekends (convoluted test. hmm..)
		eom += timedelta(days=1)

	hols = TradingHolidays()
	s.every("eom-businessday", calendar=hols).do(job, x="hello", y="eom")
	eom = dt.fromtimestamp(s.jobs[2].next_timestamp)
	assert(eom.isoweekday() < 6 and eom not in hols)  # make sure it's a businessday
	eom += timedelta(days=1)
	while eom.day != 1:  # if there are any days between eom-businessday and 1st of next month, those have to be weekends or holidays (convoluted test. hmm..)
		assert(eom.isoweekday() >= 6 or eom in hols)
		eom += timedelta(days=1)


def test_monthly():
	def day_suffixed_str(day):
		suffix = 'th' if 11<=day<=13 else {1:'st',2:'nd',3:'rd'}.get(day%10, 'th')
		return str(day) + suffix

	today = dt.now().replace(hour=23, minute=59, second=0, microsecond=0)
	yesterday = (dt.now() - timedelta(days=1)).replace(hour=23, minute=59, second=0, microsecond=0)
	s = TaskScheduler()

	j = s.every(day_suffixed_str(today.day)).strict_date(True).at("23:59").do(job, x="hello", y="world")
	assert(j.next_timestamp==dt.timestamp(today))

	# yesterday's date is scheduled to next month
	j = s.every(day_suffixed_str(yesterday.day)).strict_date(False).at("23:59").do(job, x="hello", y="world")
	assert(j.next_timestamp==dt.timestamp(yesterday+monthdelta(1)))

	# strict_date is False. So 31st will just be the last day of the month
	j = s.every("31st").strict_date(False).at("23:59").do(job, x="hello", y="world")
	eom = (dt.now() + monthdelta(1)).replace(day=1, hour=23, minute=59, second=0, microsecond=0) - timedelta(days=1)
	assert(j.next_timestamp==dt.timestamp(eom))

	# strict_date is True. So next schedule will be when month has 31st
	j = s.every("31st").strict_date(True).at("23:59").do(job, x="hello", y="world")
	day = dt.now()
	while ((day + monthdelta(1)).replace(day=1) - timedelta(days=1)).day != 31:
		day += monthdelta(1)
	assert(j.next_timestamp==dt.timestamp(day.replace(day=31, hour=23, minute=59, second=0, microsecond=0)))

	with pytest.raises(BadScheduleError):
		# error because .strict_date() is not called
		s.every("1st").at("23:59").do(job, x="hello", y="world")

	assert (len(s.jobs) == 4)


def test_repeat():
	d = time.time()
	sleep_time = 1
	s = TaskScheduler()
	s.every(sleep_time).do(job, x="hello", y="world")
	assert (abs(s.jobs[0].next_timestamp - (d+sleep_time)) < 0.1)
	time.sleep(sleep_time+0.5)
	s.check()
	assert (abs(s.jobs[0].next_timestamp - (d+(2*sleep_time))) < 0.1)


def test_repeat_parallel():
	sleep_time = 1
	s = TaskScheduler()
	s.every(sleep_time).do_parallel(job, x="hello", y="world")
	s.every(sleep_time).do_parallel(job, x="hello", y="world")
	ts = s.jobs[0].next_timestamp
	d = time.time()
	assert (abs(ts - (d+sleep_time)) < 0.1)
	time.sleep(sleep_time+0.5)
	s.check()
	assert (s.jobs[0].next_timestamp == ts) # still not rescheduled
	time.sleep(0.5)
	assert (s.jobs[0].next_timestamp != ts) # rescheduled parallely
	assert (abs(s.jobs[0].next_timestamp - (d+(2*sleep_time))) < 0.1)
	assert (abs(s.jobs[0].next_timestamp - s.jobs[1].next_timestamp) < 0.1)


def test_parallel_wait_running():
	'''long running parallel jobs shouldn't step on it's own feet (don't start when another instance is running)'''
	run_count = 0
	run_count = 0
	def _slow_job():
		nonlocal run_count
		time.sleep(2)
		run_count += 1

	s = TaskScheduler()
	s.every(1).do_parallel(_slow_job)
	s.every(1).do_parallel(_slow_job)
	time.sleep(1)
	s.check() # both kick off
	time.sleep(2)
	s.check() # nothing runs here as both as running
	time.sleep(1)
	assert(run_count==2)
	s.check() # both kick off again
	time.sleep(3)
	assert(run_count==4)


def test_parallel_stopper():
	def job(x, y):
		time.sleep(2)
		print(x, y)

	s = TaskScheduler(check_interval=1)
	s.every(1).do_parallel(job, x="hello", y="world")

	def stopp():
		# job will start in 1 second, and finish in 3 seconds. Attempting to stop at the 2 second mark
		time.sleep(2)
		print("stopping thread")
		s.stop()
	t = threading.Thread(target=stopp)
	t.start()
	s.start()
	assert(any([j.is_running for j in s.jobs])==False) # successfully stopped all parallel tasks


def test_error_callback():
	sleep_time = 1
	errors = []
	err_count = 0

	def failing_job(msg):
		raise Exception(msg)

	def err(e):
		nonlocal errors, err_count
		cause = str(e).strip().split()[-1] # get last word from traceback
		errors.append(cause)
		err_count += 1

	def err_specific(e):
		nonlocal errors, err_count
		cause = str(e).strip().split()[-1] # get last word from traceback
		errors.append(cause+"_specific")
		err_count += 1

	s = TaskScheduler(on_job_error=err)
	s.every(sleep_time).do_parallel(failing_job, msg='one')
	s.every(sleep_time).do(failing_job, msg='two')
	s.every(sleep_time).do_parallel(failing_job, msg='three').catch(err_specific)
	time.sleep(sleep_time+0.5)
	s.check()
	time.sleep(0.5)
	assert(sorted(errors)==sorted(['one', 'two', 'three_specific'])) # err callbacks were called
	assert(err_count==3)
	assert(s.jobs[0].did_fail()==True)
	pretty_print(s.jobs[0].to_dict())


def test_print_capture():
	def slow_job(sleep_time):
		time.sleep(sleep_time)
		print("Slow job completed")

	sleep_time = 1
	s = TaskScheduler(log_filepath=LOGGING_TEST_FILE)
	s.every(1).do_parallel(slow_job, sleep_time=sleep_time)
	s.check()

	counter = 4
	while counter>0:
		print("outside")
		counter -= 1
		print('running:', s.jobs[0].is_running)
		s.check()
		time.sleep(0.5)
	print("stopping")
	s.join()
	j0 = s.jobs[0].to_dict()
	assert('Slow job completed' in j0['logs']['log'])
	assert('===' in j0['logs']['log'])
	assert('outside' not in j0['logs']['log'])
	assert('stopping' not in j0['logs']['log'])
	pretty_print(j0)
	# test log file
	assert(os.path.isfile(LOGGING_TEST_FILE)==True)
	with open(LOGGING_TEST_FILE, 'r') as lf:
		assert('Slow job completed' in lf.read())



def test_log_rotation():
	def slow_job(sleep_time):
		time.sleep(sleep_time)
		print("Slow job completed")

	sleep_time = 1
	log_size_limit = 200 # NOTE: this has to be longer than 1 line at least
	s = TaskScheduler(log_filepath=LOGGING_TEST_FILE, log_maxsize=log_size_limit)
	s.every(1).do_parallel(slow_job, sleep_time=sleep_time)
	s.check()

	counter = 4
	while counter>0:
		s.check()
		counter -= 1
		time.sleep(0.5)
	print("stopping")
	s.join()

	# test log file
	assert(os.path.isfile(LOGGING_TEST_FILE)==True)
	with open(LOGGING_TEST_FILE, 'r') as lf:
		assert(len(lf.read().encode('utf-8'))<=log_size_limit)

	assert(os.path.isfile(LOGGING_TEST_FILE+".1")==True)
	with open(LOGGING_TEST_FILE+".1", 'r') as lf:
		assert(len(lf.read().encode('utf-8'))<=log_size_limit)



def test_silent_run():
	def slow_job(sleep_time):
		time.sleep(sleep_time)
		print("Slow job completed")

	sleep_time = 1
	s = TaskScheduler(log_filepath=LOGGING_TEST_FILE)
	s.every(1).do_parallel(slow_job, sleep_time=sleep_time).silently()
	s.check()

	counter = 4
	while counter>0:
		print("outside")
		counter -= 1
		print('running:', s.jobs[0].is_running)
		s.check()
		time.sleep(0.5)
	print("stopping")
	s.join()
	j0 = s.jobs[0].to_dict()
	assert('Slow job completed' in j0['logs']['log'])
	assert('===' not in j0['logs']['log'])      # '===' is printed only if job is not silent
	assert('outside' not in j0['logs']['log'])
	assert('stopping' not in j0['logs']['log'])
	pretty_print(j0)
	# test log file
	assert(os.path.isfile(LOGGING_TEST_FILE)==True)
	with open(LOGGING_TEST_FILE, 'r') as lf:
		assert('Slow job completed' in lf.read())



def test_job_docstring():
	def job_with_descr():
		'''job test docsting'''
		print("docstring test")

	def job_without_descr():
		print("no docstring test")

	s = TaskScheduler()
	j_w_descr = s.every(1).do(job_with_descr)
	j_wo_descr = s.every(1).do(job_without_descr)

	assert(j_w_descr.to_dict()['doc']=='job test docsting')
	assert(j_wo_descr.to_dict()['doc']==None)
	pretty_print(j_w_descr.to_dict())


def test_job_rerun():
	now = dt.now()
	today_str = now.strftime("%A").lower() # day of the week
	in2sec_str = now.strftime("%H:%M")
	s = TaskScheduler()
	s.every(today_str).at(in2sec_str).do(job, x="hello", y=today_str)
	assert(len(s.jobs) == 1)
	time.sleep(0.5)
	s.check() # will run here, and reschedule to next week
	j = s.jobs[0]
	run_end = j.to_dict()['logs']['end']
	# test if next run greater than 6 days from now
	test_timestamp = time.time()
	assert(j.next_timestamp > test_timestamp+(6*24*60*60)) # assert that next scheduled job happens after a week
	time.sleep(1)

	# rerun the job
	prev_resched_timestamp = j.next_timestamp
	s.rerun(j.jobid)
	time.sleep(1)
	rerun_end = j.to_dict()['logs']['end']
	assert(run_end != rerun_end)
	# rerun should not reschedule the job
	assert(j.next_timestamp == prev_resched_timestamp)
	assert(f"hello {today_str}" in j._run_info.log)

	# rerun the job with different args [8/5/2024]
	s.rerun(j.jobid, kwargs={'y': 'potato'})
	time.sleep(1)
	assert('potato' in j._run_info.log)
	assert(f"hello {today_str}" not in j._run_info.log) # This does not exist because today_str was overridden with 'potato'

	s.rerun(j.jobid, kwargs={'y': 'potato', 'z': 'nonsense'})
	time.sleep(1)
	assert("unexpected keyword argument 'z'" in j._run_info.error)

	s.rerun(j.jobid)
	time.sleep(1)
	assert('potato' not in j._run_info.log)
	assert(f"hello {today_str}" in j._run_info.log)




def test_job_disable():
	s = TaskScheduler()
	j = s.every(1).do(job, x="hello", y="state")
	time.sleep(1)
	s.check()
	data = j.to_dict()
	first_run_start = data['logs']['start']
	assert(data['is_disabled']==False)
	assert(first_run_start is not None) # runs fine

	# test job level disable/enable
	j.disable()
	time.sleep(1)
	s.check()
	data = j.to_dict()
	assert(data['is_disabled']==True)
	assert(data['logs']['start'] == first_run_start) # did not run again as it was disabled

	j.enable()
	time.sleep(1)
	s.check()
	data = j.to_dict()
	assert(data['is_disabled']==False)
	assert(data['logs']['start'] > first_run_start) # ran again after enabling

	# test scheduler level disable/enable
	latest_run_start = data['logs']['start']
	s.disable_all()
	time.sleep(1)
	s.check()
	data = j.to_dict()
	assert(data['is_disabled']==True)
	assert(data['logs']['start'] == latest_run_start) # did not run again as it was disabled

	s.enable_all()
	time.sleep(1)
	s.check()
	data = j.to_dict()
	assert(data['is_disabled']==False)
	assert(data['logs']['start'] > latest_run_start) # ran again after enabling




@pytest.mark.filterwarnings("ignore:I/O error")
def test_timezones():
	tomorrow = dt.now() + timedelta(days=1)
	tomorrow_str = tomorrow.strftime("%A").lower() # day of the week
	in2sec_str = tomorrow.strftime("%H:%M")

	s = TaskScheduler()
	with pytest.raises(BadScheduleError):
		s.every(tomorrow_str).at("8:00").timezone("US/US").do(job, x="hello", y=today_str) # bad timezone

	euro = s.every(tomorrow_str).at(in2sec_str).timezone("Europe/London").do(job, x="hello", y=tomorrow.strftime("%A"))
	assert(len(s.jobs) == 1)
	local = s.every(tomorrow_str).at(in2sec_str).timezone("America/New_York").do(job, x="hello", y=tomorrow.strftime("%A"))
	assert((euro.next_timestamp - local.next_timestamp)/60/60 in (-4, -5))

	# test if next run greater than 6 days from now
	# 'now' is in NY time and schedule is in London. So it will automatically be rescheduled to next run
	now = dt.now(tz.gettz("America/New_York"))
	today_str = now.strftime("%A").lower() # day of the week
	in2sec_str = now.strftime("%H:%M")
	s.every(today_str).at(in2sec_str).timezone("Europe/London").do(job, x="hello", y=today_str)
	test_timestamp = now.timestamp()
	assert(s.jobs[-1].next_timestamp > test_timestamp+(6*24*60*60))
	time.sleep(1)

	est = s.on(dt(now.year+1, 12, 1).strftime("%Y-%m-%d")).at("8:00").timezone("America/New_York").do(job, x="hello", y=today_str)
	# gmt = s.on(dt(now.year+1, 12, 1).strftime("%Y-%m-%d")).at("8:00").timezone("Europe/London").do(job, x="hello", y=today_str)
	assert(est.to_datetime(est.next_timestamp).strftime('%Z')=="EST")

	edt = s.on(dt(now.year+1, 6, 1).strftime("%Y-%m-%d")).at("8:00").timezone("America/New_York").do(job, x="hello", y=today_str)
	assert(edt.to_datetime(edt.next_timestamp).strftime('%Z')=="EDT")



def test_fs_persistent_logs():
	s = TaskScheduler() # persist_states=True by default
	j1 = s.every(1).do_parallel(job, x="hello", y="state1")
	j2 = s.every(1).do(job, x="hello", y="state2") # make argument slightly different so that job signature is different
	time.sleep(2)
	s.check()
	time.sleep(1)

	jobs_state_dir = s._state_handler._job_state_dir
	for j in [j1, j2]:
		state_file = os.path.join(jobs_state_dir ,f"{j.signature_hash()}.pickle")
		assert(os.path.isfile(state_file))

		import pickle
		with open(state_file, 'rb') as f:
			state = pickle.load(f)
			data = state['logs']
		assert(isinstance(data['start'], dt))
		assert(isinstance(data['end'], dt))
		assert(state['disabled']==False)

		assert(j._run_info._ended_at==data['end'])

	s = TaskScheduler(persist_states=False)
	j = s.every(1).do(job, x="hello", y="state")
	time.sleep(1)
	s.check()
	assert(s._state_handler is None)
	assert(isinstance(j._run_info._ended_at, dt)) # test if it ran even without persist_states


@pytest.fixture
def sqlalchemy_state():
	state = SQLAlchemyState(f"sqlite:///{DB_STATE_TEST_FILE}")
	yield state
	state._engine.dispose()
	if os.path.isfile(DB_STATE_TEST_FILE):
		os.remove(DB_STATE_TEST_FILE)


def test_sqlite_persistent_logs(sqlalchemy_state):
	assert(not os.path.isfile(DB_STATE_TEST_FILE)) # file should not be created until first use

	s = TaskScheduler(state_handler=sqlalchemy_state) # persist_states=True by default
	assert(s._state_handler is not None)

	j = s.every(1).do(job, x="hello", y="state")
	s.every(1).do_parallel(job, x="hello", y="state2") # make argument slightly different so that job signature is different
	time.sleep(2)

	assert(j._run_info._ended_at is None)
	assert(not os.path.isfile(DB_STATE_TEST_FILE)) # file should not be created until first use

	s.check()
	time.sleep(1)

	assert(os.path.isfile(DB_STATE_TEST_FILE)) # file saved at the s.check() call
	assert(j._run_info._ended_at is not None)
	assert(isinstance(j._run_info._ended_at, dt)) # test if state was restored


	# restore saved states
	s = TaskScheduler(state_handler=sqlalchemy_state)
	j = s.every(1).do(job, x="hello", y="state")

	s.restore_all_job_logs() # need to manually call restore method - will be automatically called if s.start() is used

	assert(j._run_info._ended_at is not None)
	assert(isinstance(j._run_info._ended_at, dt))

	time.sleep(1)
	s.check()
	assert(isinstance(j._run_info._ended_at, dt))



@pytest.fixture
def script_dir():
	dir_path = os.path.join(os.getcwd(), "testscript_dir")
	try:
		shutil.rmtree(dir_path)
	except:
		pass
	os.makedirs(dir_path)
	yield dir_path
	try:
		shutil.rmtree(dir_path)
	except:
		pass


def test_run_script(script_dir):
	# basic test setup
	wd = os.getcwd()

	script_name = "testscript.py"
	failing_script_name = "testscript_failing.py"
	import_name = "import_me"
	error_count = 0
	def _error_handler(e):
		nonlocal error_count
		error_count += 1
		print(e)
	# working directory test
	# here we need to check and make sure that the script can import from adjacent files
	# and not import files outside. ugh
	with open(os.path.join(script_dir, import_name + ".py"), 'w') as f:
		f.write("print('in import_me')\n")
		f.write("def test():\n")
		f.write("\tprint('import works')\n")

	with open(os.path.join(script_dir, script_name), 'w') as f:
		f.write(f"import time\n")
		f.write(f"import sys\n")
		f.write(f"import warnings\n")
		f.write(f"import {import_name}\n")
		f.write("print(sys.argv)\n")
		f.write(f"{import_name}.test()\n")
		f.write("time.sleep(1)\n")
		f.write("warnings.warn('warnings should not cause error')\n") #
		f.write("print('Done')\n")

	with open(os.path.join(script_dir, failing_script_name), 'w') as f:
		f.write(f"1/0\n")

	s = TaskScheduler(on_job_error=_error_handler)
	j_parallel = s.every(1).run_script_parallel(script_dir, failing_script_name)
	j_parallel2 = s.every(1).run_script_parallel(os.path.basename(script_dir), script_name) # testing relative name
	j = s.every(1).run_script(os.path.basename(script_dir), script_name) # relative script dir path
	j = s.every(1).run_script(os.path.basename(script_dir), script_name, script_args=['--test_arg']) # relative script arguments

	time.sleep(1.5)
	s.check()
	time.sleep(1)
	assert(isinstance(j_parallel._run_info._ended_at, dt))
	assert(isinstance(j_parallel2._run_info._ended_at, dt))
	assert(isinstance(j._run_info._ended_at, dt))
	assert(script_dir in j_parallel._run_info.log)
	assert(script_dir in j_parallel2._run_info.log)
	assert(script_dir in j._run_info.log)

	assert('ZeroDivisionError' in j_parallel._run_info.error)
	assert(j_parallel2._run_info.error == '')
	assert(j._run_info.error == '')

	assert(error_count == 1) # failing_script_name should be the only one that fails

	# test script logs
	assert('import works' in j._run_info.log) # test import
	assert('warning' in j._run_info.log) # test warning
	assert('--test_arg' in j._run_info.log) # test script arguments
	assert('Done' in j._run_info.log) # just the last thing - test it anyway

	assert(wd == os.getcwd())


def test_external_job_class():

	class OnePMJob(Job):

		@classmethod
		def is_valid_interval(cls, interval, time_string):
			# sending dummy time_string="00:00" to Job.is_valid_interval because it can't be None
			return interval=='1pm'

		def schedule_next_run(self, *args, **kwargs):
			d = dt.now().replace(hour=13, minute=0, second=0, microsecond=0)
			if d < dt.now():
				d = d + timedelta(days=1)

			self.next_timestamp = d.timestamp()

	s = TaskScheduler()

	with pytest.raises(BadScheduleError): # this fails if a BadScheduleError is not raised
		s.every('1pm').do(job)

	s.register_external_job_class(OnePMJob)
	j = s.every('1pm').do(job)

	assert(dt.fromtimestamp(j.next_timestamp).hour == 13)



import sys
from datetime import datetime as dt
import threading

from dateutil import tz
from contextlib import contextmanager
import traceback
import logging

from ._capture import print_capture


# default logging configuration
# logging.captureWarnings(True)
LOG_FORMATTER = logging.Formatter('%(message)s')
LOGGER_NAME = 'flask_production'
LOGGER = logging.getLogger(LOGGER_NAME)
LOGGER.setLevel(logging.INFO)
# stop propagting to root logger
LOGGER.propagate = False




class _PrintLogger(object):
	'''
	logging class to capture any print statements within a job
	also captures start time, end time and error traceback
	'''

	def __init__(self, tzname=None):
		self._lock = threading.Lock()
		self._reset()
		self._tzname = tzname

	@property
	def log(self):
		with self._lock:
			return self._run_log

	@property
	def error(self):
		with self._lock:
			return self._err_log

	@property
	def started_at(self):
		with self._lock:
			return self._started_at

	@property
	def ended_at(self):
		with self._lock:
			return self._ended_at

	def _reset(self):
		'''clear previous run info'''
		with self._lock:
			self._run_log = ''
			self._err_log = ''
			self._started_at = None
			self._ended_at = None

	def _log_callback(self, msg: str):
		'''
		writting to stderr since stdout is being redirected here. Using print() will be circular
		log to file using the logging library if LOGGER handler is set by TaskScheduler
		'''
		if msg.strip()=='':return
		msg = msg.replace('\r\n', '\n') # replace line endings to work correctly
		sys.stderr.write(msg)
		if len(LOGGER.handlers)>0:
			LOGGER.info(msg.strip())
		with self._lock:
			self._run_log += msg

	@contextmanager
	def start_capture(self):
		'''
		begin recording print statements
		'''
		self._reset() # clear previous run info
		with self._lock:
			self._started_at = dt.now(tz=tz.gettz(self._tzname))
		with print_capture(callback=self._log_callback):
			yield
		with self._lock:
			self._ended_at = dt.now(tz=tz.gettz(self._tzname))

	def set_error(self):
		'''called when job throws error'''
		with self._lock:
			self._err_log = traceback.format_exc()

	def to_dict(self):
		with self._lock:
			return dict(
				log=self._run_log,
				err=self._err_log,
				start=self._started_at,
				end=self._ended_at,
			)

	def from_dict(self, info_dict):
		if info_dict.get('start') is not None:
			with self._lock:
				self._run_log = info_dict['log']
				self._err_log = info_dict['err']
				self._started_at = info_dict['start']
				self._ended_at = info_dict['end']



'''
code derived from stdio_proxy package - https://github.com/bonprosoft/stdio_proxy

MIT License

Copyright (c) 2019 Yuki Igarashi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

import _pyio
import threading
import sys



class _ThreadSafeTextIOWrapperProxy(_pyio.TextIOWrapper):  # type: ignore
    def __init__(self, *args, **kwargs):  # type: ignore
        self._local_objects = threading.local()
        super().__init__(*args, **kwargs)

    def register(self, new_buffer, no_close=True):
        self._local_objects.buffer = new_buffer
        self._local_objects.no_close = no_close
        self._local_objects.registered = True

    def unregister(self):
        self._local_objects.buffer = None
        self._local_objects.registered = False

    @property
    def _registered(self):
        return getattr(self._local_objects, "registered", False)

    @property
    def buffer(self):
        if not self._registered:
            return self._buffer

        buf = self._local_objects.buffer
        assert buf is not None
        return buf

    def close(self):
        if self._registered and self._local_objects.no_close:
            self.buffer.flush()
        else:
            super().buffer.flush()

    def write(self, s):
        return super().write(s)


def _create_proxy(target):
    return _ThreadSafeTextIOWrapperProxy(
        target.buffer,
        target.encoding,
        getattr(target, "errors", "strict"),
        getattr(target, "newlines", None),
        getattr(target, "line_buffering", True),
    )



class _redirect_stdout(object):
    """Context manager for temporarily redirecting stdout to an object.

    Note:
        For Python 3 environments, ``dst_obj`` is supposed to handle ``bytes``, not ``str``.
        For Python 2 environments, ``str``, not ``unicode``.
        The easiest way to prepare ``dst_obj`` for both environments is to use an ``io.BytesIO`` object.

        For Python 3 environments, the interface of ``dst_obj`` should be compatible with that of ``io.BufferedWriter``.
        For instance, if you use `sys.stdin.readline`, ``dst_obj`` must have ``write`` method, ``flush`` method, and ``closed`` property.
        Note that the argument of ``write`` method is ``bytes``, not ``str``.

        For Python 2 environments, the interface of ``dst_obj`` should be compatible with that of the Python file object.

    Examples:
        >>> buf = io.BytesIO()
        >>> with redirect_stdout(buf):
        ...     sys.stdout.write("Foo\n")
        >>> print("Redirected: {}".format(buf.getvalue()))
        Redirected: b'Foo\n'

    """

    _lock = threading.Lock()
    _original = None
    _proxy = None
    _n_use = 0

    def __init__(self, new_obj):
        self._new_obj = new_obj

    def __enter__(self):
        with _redirect_stdout._lock:
            if _redirect_stdout._n_use == 0:
                if _redirect_stdout._original is None:
                    _redirect_stdout._original = getattr(sys, 'stdout')
                    _redirect_stdout._proxy = _create_proxy(_redirect_stdout._original)

                setattr(sys, 'stdout', _redirect_stdout._proxy)

            _redirect_stdout._n_use += 1
            _redirect_stdout._proxy.register(self._new_obj)

    def __exit__(self, exc_type, exc_val, exc_tb):
        _redirect_stdout._proxy.unregister()
        with _redirect_stdout._lock:
            _redirect_stdout._n_use -= 1

            if _redirect_stdout._n_use == 0:
                setattr(sys, 'stdout', _redirect_stdout._original)


'''
implementing print statement capture using the above thread safe stdout proxy
'''
# from stdio_proxy import redirect_stdout

class _PrintCapture(_pyio.BytesIO):
    '''File-like object that will be the destination of capture'''
    def __init__(self, callback):
        self._buf = b''
        self._write_cb = callback

    def write(self, b):
        self._buf += b
        if b.endswith(b'\n'):
            line = self._buf.decode(errors='ignore')
            self._write_cb(line)
            self._buf = b''

    def capture(self):
        return _redirect_stdout(self)


def print_capture(callback):
    return _PrintCapture(callback).capture()



from datetime import datetime as dt
from collections import OrderedDict
import json
import random
import string
import inspect

from dateutil import tz
from flask import Flask, Blueprint, request, send_file


from .html_templates import * # pylint: disable=unused-wildcard-import



class TaskMonitor(object):
	'''
	Web interface to monitor and manage tasks

	Args:
	- app (int): ``Flask`` application
	- sched (TaskScheduler): task scheduler with task definitions
	- display_name (str): name of the application to be displayed
		- default app.name

	- endpoint (str): URL endpoint where the taskmonitor can be viewed
		- default "@taskmonitor"
	- homepage_refresh (int): home page auto refresh interval (in seconds)
		- default 30
	- taskpage_refresh (int): task page auto refresh interval (in seconds)
		- default 5
	- can_rerun (bool): if True adds `rerun` button to job page
		- default True
	- can_disable (bool): if True adds `disable` button to job page
		- default True
	'''
	def __init__(self, app:Flask, sched, display_name=None, endpoint="@taskmonitor", homepage_refresh=30, taskpage_refresh=5,
		can_rerun=True, # adds rerun button to job page
		can_disable=True, # adds disable button to job page
		):
		self.tzname = sched._tz_default
		self._init_dt = dt.now(tz.gettz(self.tzname)).strftime("%m/%d/%Y %I:%M %p %Z") # preformatted start time
		self.app = app
		self.sched = sched
		self._endpoint = endpoint
		self._display_name = display_name or self.app.name
		self._homepage_refresh = homepage_refresh
		self._taskpage_refresh = taskpage_refresh
		self._api_protection_token = ''.join(random.choices(string.ascii_letters, k=20)) # protect job reruns external API calls. makes sure endpoint can only be called by TaskMonitor

		self._can_rerun = can_rerun
		self._can_disable = can_disable

		bp = Blueprint('taskmonitor_bp', __name__, url_prefix=f"/{self._endpoint}")

		bp.add_url_rule("", view_func=self.__show_all, methods=['GET'])
		bp.add_url_rule("/<int:n>", view_func=self.__show_one, methods=['GET'])
		bp.add_url_rule("/rerun", view_func=self.__rerun_job, methods=['POST'])
		bp.add_url_rule("/enable_disable", view_func=self.__enable_disable_job, methods=['POST'])
		bp.add_url_rule("/json/all", view_func=self.__get_all_json, methods=['GET'])
		bp.add_url_rule("/json/summary", view_func=self.__get_summary_json, methods=['GET'])
		bp.add_url_rule("/json/<int:n>", view_func=self.__get_one_json, methods=['GET'])

		bp.add_url_rule("/static/<type>/<filename>", view_func=self.__serve_file, methods=['GET'])
		self.app.register_blueprint(bp)

		_favicon_is_set = False
		for rule in self.app.url_map.iter_rules():
			if 'favicon.ico' in str(rule):
				_favicon_is_set = True

		if not _favicon_is_set:
			self.app.add_url_rule("/favicon.ico", view_func=lambda: self.__serve_file('ico', 'flask_boiler.ico'), methods=['GET'])


	def __html_wrap(self, elems, css=[]):
		return HTML(''.join(elems), title="{} Task Monitor".format(self._display_name), css=css)

	def __js_src_wrap(self, filename):
		return SCRIPT_SRC(f'/{self._endpoint}/static/js/{filename}')

	def __css_src_wrap(self, filename):
		return STYLE_LINK(f'/{self._endpoint}/static/css/{filename}')

	def __serve_file(self, type, filename):
		if 'max_age' in inspect.getfullargspec(send_file).args:
			return send_file(os.path.join(WEB_FOLDER, type, filename), max_age=86400)
		else:
			return send_file(os.path.join(WEB_FOLDER, type, filename), cache_timeout=86400)


	def __state(self, jdict):
		state = {'state':'READY', 'css': 'grey', 'title': '' }
		if jdict['is_disabled']:
			state['state'] = "DISABLED"
			state['css'] = "blue"
			return state # no need to check status if disabled

		if jdict['is_running']:
			state['state'] = "RUNNING"
			state['css'] = "yellow"
		elif jdict['logs']['err'].strip()!='':
			state['state'] = "ERROR"
			state['css'] = "red"
			state['title'] = jdict['logs']['err'].strip().split("\n")[-1]
		elif jdict['logs']['end'] is not None and jdict['logs']['log'].strip()!='':
			state['state'] = "SUCCESS"
			state['css'] = "green"
		return state

	def __duration(self, jdict):
		duration = None
		if jdict['logs']['start'] is not None and jdict['logs']['end'] is not None:
			seconds = (jdict['logs']['end']-jdict['logs']['start']).seconds
			if seconds >= 60:
				minutes = seconds // 60
				seconds = seconds % 60
				duration = "{}:{} minutes".format(minutes, str(seconds).zfill(2))
			elif seconds==1:
				duration = "{} second".format(seconds)
			else:
				duration = "{} seconds".format(seconds)
		return duration

	def __timestr_to_12hr(self, tstr):
		d = dt.strptime(dt.now().strftime("%Y-%m-%d "+ str(tstr)), "%Y-%m-%d %H:%M")
		return d.strftime("%I:%M%p")


	def __scheduleTD(self, jdict):
		tz_str = ''
		if isinstance(jdict['tzname'], str):
			tz_str = dt.now(tz.gettz(jdict['tzname'])).strftime("[%Z]")

		if isinstance(jdict['every'], int): # jdict['type']=='RepeatJob'
			out = "every {} seconds {}".format(jdict['every'], tz_str)
			return TD(out.strip())

		elif jdict['type']=='OneTimeJob':
			out = "on {} at {} {}".format(jdict['every'], jdict['at'], tz_str)
			return TD(out.strip())

		elif jdict['type']=='NeverJob':
			out = 'on-demand'
			return TD(out)

		elif isinstance(jdict['at'], (list,set,tuple)):
			full_str = "every {} at {} {}".format(jdict['every'], ', '.join(jdict['at']), tz_str)

			if len(jdict['at']) >= 5:
				at_str = ', '.join(jdict['at'][:3]) + ', ...' + jdict['at'][-1]
				out = "every {} at {} {}".format(jdict['every'], at_str, tz_str)
				return TD(out.strip(), attrs={'title': full_str.strip()})
			else:
				return TD(full_str.strip())
		else:
			out = "every {} at {} {}".format(jdict['every'], jdict['at'], tz_str)
			return TD(out.strip())


	def __date_fmt(self, d):
		fallback = '-'+('&nbsp;'*30) # a hiphen and some html spaces
		return d.strftime("%Y-%m-%d %H:%M:%S %Z") if d is not None else fallback

	def __date_sort_attr(self, d):
		return {'data-sort': d.timestamp() if d is not None else 0}

	def __descrTD(self, d):
		if d is None: return TD('-')
		d = d.strip()
		short_d = d[:30] + "..."
		return TD(short_d if len(d)>30 else d, attrs={'title': d})

	def __src_err_line(self, j):
		if j['logs']['err'].strip()!='':
			for l in j['logs']['err'].strip().split("\n"):
				if l.strip():
					idx = j['src'].find(l.strip())
					if idx>=0:
						return j['src'][:idx].count("\n")
		return -1

	def __get_all_json(self):
		if len(self.sched.jobs)==0:
			return json.dumps({'error':'Nothing here'})
		else:
			return json.dumps({'success': [j.to_dict() for j in self.sched.jobs]}, default=str)

	def __get_summary_json(self):
		if len(self.sched.jobs)==0:
			return json.dumps({'error':'Nothing here'})
		else:
			details = []
			summary = {'count': 0, 'running': 0, 'errors': 0}
			for j in self.sched.jobs:
				jd = j.to_dict()
				state = self.__state(jd)
				summary['count'] += 1
				if state['state'] == "ERROR":
					summary['errors'] += 1
				elif state['state'] == 'RUNNING':
					summary['running'] += 1
				details.append({
					'id': jd['jobid'],
					'state': state['state'],
					'signature': jd['signature'],
					'prev_run': jd['logs']['start'],
					'next_run': jd['next_run'],
				})
			out = {'name': self._display_name, 'summary': summary, 'details': details}
			return json.dumps({'success': out}, default=str)

	def __get_one_json(self, n):
		j = self.sched.get_job_by_id(n)
		if j is None:
			return json.dumps({'error':'Invalid job id'})
		return json.dumps({'success': j.to_dict()}, default=str)

	def __show_all(self):
		if len(self.sched.jobs)==0:
			return 'Nothing here'
		d = []
		for j in self.sched.jobs:
			jd = j.to_dict()
			duration = self.__duration(jd)
			state = self.__state(jd)
			start_dt = jd['logs']['start']
			end_dt = jd['logs']['end']
			next_dt = jd['next_run']
			next_dt_str = ''
			if next_dt is None:
				if jd['is_disabled']:
					next_dt_str = 'Disabled'
				else:
					next_dt_str = 'Never'
			else:
				next_dt_str = self.__date_fmt(next_dt)

			d.append(OrderedDict({
				'Id': TD(jd['jobid']),
				'Name': TD(jd['func'].replace('<', '&lt;').replace('>', '&gt;'), attrs={'title':j.func_signature()}),
				'Schedule': self.__scheduleTD(jd),
				'Description': self.__descrTD(jd['doc']),
				'State': TD(state['state'], css=state['css'], attrs={'title': state['title']}),
				'Start': TD(self.__date_fmt(start_dt), attrs=self.__date_sort_attr(start_dt)),
				'End': TD(self.__date_fmt(end_dt), attrs=self.__date_sort_attr(end_dt)),
				'Time Taken': TD(duration),
				'Next Run': TD(next_dt_str, attrs=self.__date_sort_attr(next_dt)),
				'More':TD("<a href='/{}/{}'><button>show more</button><a>".format(self._endpoint, jd['jobid']))
			}))
		rows = [TR(row.values()) for row in d]
		head = [TH(th, default_sort=(th=="Next Run") ) for th in d[0].keys()]	# apply sorting to 'next run'
		all_jobs_table = TABLE(thead=THEAD(head), tbody=TBODY(rows), elem_id='all-jobs', css='all-jobs')
		refresh_text = SMALL(f"Auto-refresh in {SPAN(self._homepage_refresh, attrs={'id': 'refresh-msg'})} seconds")
		filter_input = INPUT("", attrs={'type':'text', 'placeholder':'Filter', 'id': 'filter-box'})

		js_auto_reload_variables = '''let COUNT_DOWN = {};'''.format(self._homepage_refresh)

		container = DIV(
			'\n'.join([
				H(2, "{} - Task Monitor".format(self._display_name)),
				SPAN("Running since {}".format(self._init_dt)),
				refresh_text,
				filter_input,
				all_jobs_table,
			]),
			css=["container", "container-vertical", 'center']
		)

		return self.__html_wrap(
			[
				self.__css_src_wrap('dark_theme.css'),
				self.__css_src_wrap('taskmonitor.css'),
				container,
				SCRIPT(js_auto_reload_variables),
				self.__js_src_wrap('taskmonitor.js')
			]
		)

	def __show_one(self, n):
		j = self.sched.get_job_by_id(n)
		if j is None:
			return 'Not found'
		jobd = j.to_dict()
		titleTD = lambda t: TD(t, css='title')
		state = self.__state(jobd)
		job_funcname = jobd['func'].replace('<', '&lt;').replace('>', '&gt;')

		enable_disable_btn = '''<button class="btn enable-disable-btn"
									onclick="enable_disable('{name}', {jobid}, {job_disable})"
									{btn_disabled}>
								{btn_name}
								</button>'''.format(
			name=job_funcname, jobid=n,
			job_disable="true" if not jobd['is_disabled'] else "false",
			btn_disabled="disabled" if state['state']=="RUNNING" else "",
			btn_name="Disable" if not jobd['is_disabled'] else "Enable",
		)
		rerun_btn = '''<button class="btn rerun-btn"
							onclick="rerun_trigger('{name}', {jobid})"
							{btn_disabled}>
						Rerun
						</button>'''.format(
			name=job_funcname, jobid=n, # rerun_trigger params
			btn_disabled="disabled" if state['state']=="RUNNING" or jobd['is_disabled'] else ""
		)
		rows = [
			TR([ titleTD("Schedule"), self.__scheduleTD(jobd), ]),
			TR([ titleTD("State"), TD(state['state'], css=state['css']) ]),
			TR([ titleTD("Start Time"), TD(self.__date_fmt(jobd['logs']['start'])) ]),
			TR([ titleTD("End Time"), TD(self.__date_fmt(jobd['logs']['end'])) ]),
			TR([ titleTD("Time Taken"), TD(self.__duration(jobd)) ]),
			TR([ titleTD("Next Run In"), TD("-", attrs={'id':'next-run-in'}) ]),
			TR([ TD(enable_disable_btn, colspan=2, css=['monitor-btn']) ]) if self._can_disable else '',
			TR([ TD(rerun_btn, colspan=2, css=['monitor-btn']) ]) if self._can_rerun else ''
		]

		info_table = TABLE(tbody=TBODY(rows), css='info_table')
		description_div = DIV( CODE(jobd['src'], css='python'), css=['console-color', 'console-div', 'brdr', 'monitor-code'])
		title = H(2, job_funcname, attrs={'title': j.func_signature()})
		monitor_div = DIV(
			title + info_table + description_div,
			css="monitor"
		)

		logs_row = TR([
			TD( DIV( CODE(jobd['logs']['log'], css='accesslog'), css='console-div'), css="console-color"),
			TD( DIV( CODE(jobd['logs']['err'], css='accesslog'), css='console-div'), css="console-color"),
		])
		logs_table = TABLE(thead=THEAD([TH('Logs'), TH('Traceback')]), tbody=TBODY(logs_row), css='log_table')
		logs_div = DIV( logs_table, css="logs_div" )

		container = DIV(
			monitor_div + logs_div,
			css="container"
		)

		next_run_ts = '"Never"'
		if jobd['is_disabled']:
			next_run_ts = '"Disabled"'
		elif jobd['next_run']:
			next_run_ts = jobd['next_run'].timestamp()

		variables_script = f'''
		let RUNNING = {int(jobd['is_running'])};
		let NEXT_RUN = {next_run_ts};
		let ERR_LINE = {self.__src_err_line(jobd)};
		let TASKPAGE_REFRESH = {self._taskpage_refresh};
		let API_TOKEN = '{self._api_protection_token}';
		'''

		return self.__html_wrap(
			[
				self.__css_src_wrap('dark_theme.css'),
				self.__css_src_wrap('taskmonitor.css'),
				container,
				SCRIPT(variables_script),
				self.__js_src_wrap('task.js')
			]
		)


	def __rerun_job(self):
		error = None
		data = json.loads(request.data)
		print("> rerun", data)
		if 'api_token' not in data or data['api_token']!=self._api_protection_token:
			error = 'Invalid token. Rerun blocked. Please reload the page and try again'

		elif 'jobid' not in data or not isinstance(data['jobid'], int):
			error = 'Invalid input'
		else:
			try:
				self.sched.rerun(data['jobid'])
			except Exception as e:
				error = str(e)

		if error is not None:
			return json.dumps({'error': error})
		else:
			return json.dumps({'success': True})


	def __enable_disable_job(self):
		error = None
		data = json.loads(request.data)
		print("> enable_disable", data)
		if 'api_token' not in data or data['api_token']!=self._api_protection_token:
			error = 'Action blocked'

		if 'jobid' not in data or not isinstance(data['jobid'], int):
			error = 'Invalid input'
		else:
			try:
				j = self.sched.get_job_by_id(data['jobid'])
				if j is None:
					raise ValueError("Job not found")
				if data['disable']==True:
					j.disable()
				else:
					j.enable()
			except Exception as e:
				error = str(e)

		if error is not None:
			return json.dumps({'error': error})
		else:
			return json.dumps({'success': True})



import os, sys
import subprocess
from types import ModuleType
from typing import List


class ScriptFunc(ModuleType):

    def __init__(self, script_dir_path: str, script_name: str, script_args: List[str]=[]) -> None:
        script_dir_path = os.path.abspath(script_dir_path)
        if not os.path.isdir(script_dir_path):
            raise ValueError(f"'{script_dir_path}' not found")

        if not script_name.endswith(".py"):
            raise ValueError("Only python scripts supported at this time")

        if not isinstance(script_args, list):
            raise TypeError('script_args should be a list')

        self.__file__ = os.path.join(script_dir_path, script_name)

        if not os.path.isfile(self.__file__):
            raise FileNotFoundError(f"'{self.__file__}' not found")

        self.script_dir_path = script_dir_path
        self.script_name = script_name
        self.script_args = list(str(s) for s in script_args)

        self.__module__ = self.script_dir_path
        self.__qualname__ = f'[{self.script_name}]'
        if len(self.script_args) > 0:
            self.__qualname__ += ' ' + ' '.join(self.script_args)

        self.__doc__ = f"Script called '{self.script_name}' defined in '{self.script_dir_path}'"

        self.__wd = os.getcwd() # capture working directory to change back to once script is complete


    def __call__(self):
        os.chdir(self.script_dir_path)
        cmd = [sys.executable, "-u", self.__file__] + self.script_args

        try:
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            # Grab stdout line by line as it becomes available.  This will loop until p terminates.
            while p.poll() is None:
                l = p.stdout.readline().decode().strip() # This blocks until it receives a newline.
                print(l)
            # # When the subprocess terminates there might be unconsumed output
            # # that still needs to be processed.
            print(p.stdout.read().decode().strip())
            err = p.stderr.read().decode().strip()
            if p.returncode != 0:
                e = err.split('\n')[-1]
                raise Exception(f"{e}\n\nraised from subprocess:\n{err}")
            else:
                print(err) # not really an error? maybe a warning, or script returns 0 after failure :/
            p.wait()

        finally:
            os.chdir(self.__wd)




import os
import pickle

from .base import BaseStateHandler




class FileSystemState(BaseStateHandler):

	def __init__(self, uri=None) -> None:
		super().__init__()
		if uri is None:
			uri = os.path.join(self._get_current_app_data_directory(), "states")

		self._job_state_dir = uri
		if not os.path.isdir(self._job_state_dir):
			os.makedirs(self._job_state_dir)


	def _get_current_app_data_directory(self):
		# create the unique data directory for current app
		# this folder can contain
		# - info file that includes data used to come up with current app signature
		# - job state information
		cur_app_data_dir_path = os.path.join(
			os.environ.get('APPDATA') or
			os.environ.get('XDG_DATA_HOME') or
			os.path.join(os.environ['HOME'], '.local', 'share'),
			"flask_production_data",
			self._cur_app_unique_info_hash
		)
		if not os.path.isdir(cur_app_data_dir_path):
			os.makedirs(cur_app_data_dir_path)

		# create an *.cwd file that is named after the current working directory name
		# - helps to easily identify the app and manually inspect files in the data folder
		# - contains _cur_app_unique_info used to come up with the self._cur_app_unique_info_hash
		with open(os.path.join(cur_app_data_dir_path, f'{os.path.basename(os.getcwd())}.cwd'), 'w') as f:
			for info in self._cur_app_unique_info:
				f.write(str(info)+"\n")

		return cur_app_data_dir_path


	def save_job_logs(self, job_obj):
		if self._job_state_dir is not None:
			filename = job_obj.signature_hash()
			with open(os.path.join(self._job_state_dir, f"{filename}.pickle"), 'wb') as f:
				logs = job_obj._logs_to_dict()
				pickle.dump({'logs':logs, 'disabled': job_obj.is_disabled}, f) # we only care about logs


	def restore_all_job_logs(self, jobs_list):
		if self._job_state_dir is not None:
			found_states = []
			for j in jobs_list.copy(): # work on a shallow copy of this list - safer in case the list changes. TODO: maybe use locks instead?
				filename = f"{j.signature_hash()}.pickle"
				filepath = os.path.join(self._job_state_dir, filename)
				if os.path.isfile(filepath):
					with open(filepath, 'rb') as f:
						state = pickle.load(f)
						logs = state['logs'] if 'logs' in state else state # doing it this way for backwards compatibility as 'state' was previously 'logs'
						j._logs_from_dict(logs)
						if state.get('disabled'):
							j.disable()
					found_states.append(filename)
					# print("restored", j)
			# clean up other states that did not match current jobs list (possible stale)
			for f in os.listdir(self._job_state_dir):
				if f not in found_states:
					os.remove(os.path.join(self._job_state_dir, f))

			print("* scheduler state restored from filesystem *")



import os
import holidays
from datetime import date
from dateutil.relativedelta import relativedelta, MO, FR
from dateutil.easter import easter


GOOD_FRIDAY_TRADING_HOLIDAY = str(os.environ.get('GOOD_FRIDAY_TRADING_HOLIDAY')).strip() == '1'


class TradingHolidays(holidays.countries.UnitedStates):
	def _populate(self, year):
		# Populate the holiday list with the default US holidays
		holidays.UnitedStates._populate(self, year)
		# Remove Columbus Day
		# Edit: in holidays==0.61, Columbus day was removed as a US countrywide holiday
		# 	- https://github.com/vacanza/holidays/pull/2106
		columbus_day = date(year, 10, 1) + relativedelta(weekday=MO(+2))
		if columbus_day in self:
			self.pop(columbus_day, None)

		# Remove Veterans Day
		self.pop(date(year, 11, 11), None)
		if year==2023:
			self.pop(date(year, 11, 10), None)

		# # Add Good Friday
		# NOTE: removed this since Good Friday is not a bond market holiday - [04/01/2021]
		# new NOTE: this is back to being a holiday now. So making it env variable - [4/17/2025]
		if GOOD_FRIDAY_TRADING_HOLIDAY:
			self[easter(year) + relativedelta(weekday=FR(-1))] = "Good Friday"

		# 2021-12-31 is not a holiday apparently :( and windows version seems to think it is
		nye_2021 = date(2021, 12, 31)
		if nye_2021 in self:
			self.pop(nye_2021, None)



import os
import sys
import hashlib


class BaseStateHandler:

	def __init__(self) -> None:
		# multiple apps / programs may use this library
		# _cur_app_unique_info is a way to create a unique id for each app
		# it uses:
		# - current working directory: isolates apps in the same directory from other apps on the system
		# - path to python executabe: isolates apps that use different python installations
		# - full command line:
		# 		- this includes name of the script file: isolates a script / entry point from others
		# 		- command line arguments: isolates when same script is run using different cli arguments
		self._cur_app_unique_info = [
			os.getcwd(),  		# current working directory
			sys.executable,		# python executable
			*sys.argv			# script name and all cli arguments
		]
		self._cur_app_unique_info_hash = hashlib.sha1(':'.join(self._cur_app_unique_info).encode()).hexdigest()


	def save_job_logs(self, job_obj):
		pass

	def restore_all_job_logs(self, jobs_list):
		pass



from flask import request
import cherrypy
import traceback
from datetime import datetime as dt


class CherryFlask(object):

	__slots__ = ['app', 'sched', 'timeout']

	def __init__(self, app, scheduler=None, silent=False, timeout=60):
		self.app = app
		self.sched = scheduler
		self.timeout = timeout
		if not silent:
			@app.after_request
			def _teardown(response): # pylint: disable=unused-variable
				pth = request.environ.get('PATH_INFO')
				if "@" not in pth and 'favicon' not in pth: # disable logging of endpoints like @taskmonitor, favicon.ico
					adr = request.environ.get('HTTP_X_REAL_IP', request.environ.get('REMOTE_ADDR'))
					mth = request.environ.get('REQUEST_METHOD')
					print(f'''{adr} - [{dt.now().strftime('%m/%d/%Y %H:%M:%S')}] - "{mth} {pth}" - {response.status_code}''')
				return response

	def run(self, host='0.0.0.0', port=8080, threads=5, debug=False):
		if not debug: cherrypy.config.update({'engine.autoreload.on' : False})

		cherrypy.tree.graft(self.app.wsgi_app, '/')
		cherrypy.server.unsubscribe()
		server = cherrypy._cpserver.Server()

		server.socket_host = host
		server.socket_port = port
		server.thread_pool = threads
		server.socket_timeout = self.timeout
		server.subscribe()

		if hasattr(cherrypy.engine, "signal_handler"):
			cherrypy.engine.signal_handler.subscribe()
		if hasattr(cherrypy.engine, "console_control_handler"):
			cherrypy.engine.console_control_handler.subscribe()


		try:
			cherrypy.engine.start()
			if self.sched is not None:
				cherrypy.engine.subscribe('exit', lambda:self.sched.stop())
				self.sched.start()
			else:
				cherrypy.engine.block()
		except Exception:
			traceback.print_exc()
		finally:
			self.stop()

	def stop(self):
		cherrypy.engine.exit()





from typing import Union, Callable, List
import time
from datetime import datetime as dt
from logging.handlers import RotatingFileHandler
import warnings

import holidays
from dateutil import tz

from .print_logger import (
	LOGGER,
	LOG_FORMATTER
)

from .jobs import (
	# job types
	Job,
	OneTimeJob,
	RepeatJob,
	MonthlyJob,
	AsyncJobWrapper,
	NeverJob,

	# exceptions
	BadScheduleError
)

from .script_func import ScriptFunc

from .state import (
	BaseStateHandler,
	FileSystemState,
)



USHolidays = holidays.US()


def get_local_timezone_name():
	return tz.gettz(None).tzname(dt.now())



class TaskScheduler(object):
	"""
	TaskScheduler: main class to setup, run and manage jobs

	Args:
	- check_interval (int): how often to check for pending jobs
	- holidays_calendar (holidays.HolidayBase): calendar to use for intervals like `businessday`
	- tzname (str): name of timezone as supported by dateutil.tz
	- on_job_error (function(e)): function to call if any job fail
	- log_filepath (path): file to write logs to
	- log_maxsize (int): byte limit per log file
	- log_backups (int): number of backups of logs to retain
	- startup_grace_mins (int): grace period for tasks in case a schedule was missed because of app restart
	- persist_states (bool): store job logs and read back on app restart
	- state_handler (.state.BaseStateHandler): different handler backends to store job logs
	"""

	def __init__(self,
		check_interval: int=5,
		holidays_calendar: Union[holidays.HolidayBase, None]=None,
		tzname: Union[str, None]=None,
		on_job_error: Union[Callable, None]=None,
		log_filepath: Union[str, None]=None,
		log_maxsize: int=5*1024*1024,
		log_backups: int=1,
		startup_grace_mins: int=0,
		persist_states: bool=True,
		state_handler: Union[BaseStateHandler, None]=None) -> None:

		self.jobs = []
		self._check_interval = check_interval
		self._startup_grace_mins = startup_grace_mins
		self.on_job_error = on_job_error

		tzname = tzname or get_local_timezone_name() # if None, default to local timezone
		if tz.gettz(tzname) is None:
			raise ValueError(f"unknown timezone '{tzname}'")
		self._tz_default = tzname
		print("* Default Timezone:", self._tz_default, "*")

		if holidays_calendar is not None:
			self.holidays_calendar = holidays_calendar
		else:
			self.holidays_calendar = USHolidays

		# setup logging
		self.log_filepath = log_filepath
		if self.log_filepath is not None:
			fh = RotatingFileHandler(
				filename=self.log_filepath,
				maxBytes=log_maxsize,
				backupCount=log_backups
			)
			fh.setFormatter(LOG_FORMATTER)
			LOGGER.addHandler(fh)

		# setup state persistance over app restarts
		self._state_handler = None
		if persist_states:
			self._state_handler = state_handler or FileSystemState()

		# additional job classes
		self._external_job_classes = []

		# set schedule defaults
		self.__reset_defaults()


	def __reset_defaults(self):
		self.interval = None
		self.temp_time = None
		self.tzname = self._tz_default # timezone default
		self._strict_monthly = None
		self.job_calendar = None


	def register_external_job_class(self, jclass):
		if not issubclass(jclass, Job):
			raise ValueError("class must be inherited from 'Job' class")
		self._external_job_classes.append(jclass)


	# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
	# -=-=-=-=-=-=-=-=-= New job definition methods =-=-=-=-=-=-=-=-=-
	# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

	def every(self, interval, calendar=None):
		'''
		interval is either one of the keys of RUNABLE_DAYS
		or integer denoting number of seconds for RepeatJob
		'''
		self.interval = interval
		self.job_calendar = calendar
		return self

	def on(self, *args, **kwargs):
		'''alias of .every() method'''
		return self.every(*args, **kwargs)

	def strict_date(self, strict:bool):
		'''
		required to be called when scheduling MonthlyJob
		- see MonthlyJob docstring
		'''
		if not MonthlyJob.is_valid_interval(self.interval, time_string="00:00") or not isinstance(strict, bool):
			raise BadScheduleError(".strict_date(bool) only used for monthly schedule. ex: .every('31st').strict_date(True)..")
		self._strict_monthly = strict
		return self

	def at(self, time_string):
		'''
		24 hour time string of when to run job
		- example: '15:00' for 3PM
		'''
		if self.interval is None:
			self.every('day')
		self.temp_time = time_string
		return self

	def timezone(self, tzname):
		'''
		timezone string as defined in pytz module
		- example US/Eastern
		- defaults to system timezone
		'''
		test = tz.gettz(tzname)
		if test is None:
			raise BadScheduleError(f"unknown timezone '{tzname}'")
		self.tzname = tzname
		return self

	def tz(self, *args, **kwargs):
		'''alias of .timezone() method'''
		return self.timezone(*args, **kwargs)


	def _create_job(self, func, **kwargs):
		'''
		register 'func' for the job
		- if do_parallel is True, run job in a prallel thread
		- pass kwargs into 'func' at execution
		'''
		if self.interval is None:
			raise Exception('Use .at()/.every().at() before .do()')
		if self.temp_time is None:
			self.temp_time = dt.now(tz.gettz(self.tzname)).strftime("%H:%M")

		new_jobid = len(self.jobs)
		j = None
		for jcls in self._external_job_classes:
			if jcls.is_valid_interval(self.interval, time_string=self.temp_time):
				j = jcls(new_jobid, every=self.interval, at=self.temp_time, func=func, kwargs=kwargs)
				break

		if j is None:
			if RepeatJob.is_valid_interval(self.interval, time_string=None):
				j = RepeatJob(new_jobid, every=self.interval, at=None, func=func, kwargs=kwargs)

			elif OneTimeJob.is_valid_interval(self.interval, time_string=self.temp_time):
				j = OneTimeJob(new_jobid, every=self.interval, at=self.temp_time, func=func, kwargs=kwargs)

			elif MonthlyJob.is_valid_interval(self.interval, time_string=self.temp_time):
				j = MonthlyJob(new_jobid, every=self.interval, at=self.temp_time, func=func, kwargs=kwargs, strict_date=self._strict_monthly)

			elif Job.is_valid_interval(self.interval, time_string=self.temp_time):
				j = Job(new_jobid, every=self.interval, at=self.temp_time, func=func, kwargs=kwargs)

			elif NeverJob.is_valid_interval(self.interval, time_string=None):
				j = NeverJob(new_jobid, every=self.interval, at=None, func=func, kwargs=kwargs)

		if j is None:
			raise BadScheduleError("{} is not valid\n".format(self.interval))

		j.init(
			calendar=self.holidays_calendar if self.job_calendar is None else self.job_calendar,
			tzname=self.tzname,
			generic_err_handler=self.on_job_error,
			startup_grace_mins=self._startup_grace_mins
		)
		# register callbacks to save job logs to file so it can be restored on app restart
		if isinstance(self._state_handler, BaseStateHandler):
			j.register_callback(self._state_handler.save_job_logs, cb_type="onenable")
			j.register_callback(self._state_handler.save_job_logs, cb_type="ondisable")
			j.register_callback(self._state_handler.save_job_logs, cb_type="oncomplete")

		self.__reset_defaults()
		print(j)
		return j


	def do(self, func, do_parallel=False, **kwargs):
		j = self._create_job(func, **kwargs)
		if do_parallel:
			print("================================================")
			print("==== do_parallel boolean argument will be removed")
			print("==== use do_parallel() method  instead")
			print("================================================")
			warnings.warn("do_parallel boolean argument will be removed", category=DeprecationWarning)
			j = AsyncJobWrapper(j)
		self.jobs.append(j)
		return j

	def do_parallel(self, func, **kwargs):
		'''helper function to run job in a separate thread'''
		j = self._create_job(func, **kwargs)
		j = AsyncJobWrapper(j)
		self.jobs.append(j)
		return j

	def run_script(self, script_dir_path:str, script_name:str, script_args:List[str]=[]):
		func = ScriptFunc(script_dir_path, script_name, script_args)
		j = self._create_job(func)
		self.jobs.append(j)
		return j

	def run_script_parallel(self, script_dir_path:str, script_name:str, script_args:List[str]=[]):
		func = ScriptFunc(script_dir_path, script_name, script_args)
		j = self._create_job(func)
		j = AsyncJobWrapper(j)
		self.jobs.append(j)
		return j

	# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
	# -=-=-=-=-=-=-=-= Scheduler control methods =-=-=-=-=-=-=-=-=-=-=
	# -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

	def check(self):
		'''check if a job is due'''
		for j in self.jobs.copy(): # work on a shallow copy of this list - safer in case the list changes. TODO: maybe use locks instead?
			if j.is_due() and not j.is_running:
				j.run()


	def restore_all_job_logs(self):
		try:
			if isinstance(self._state_handler, BaseStateHandler):
				self._state_handler.restore_all_job_logs(self.jobs)
		except Exception as e:
			# import traceback
			# traceback.print_exc()
			print("unable to restore states:", str(e))


	def start(self):
		'''blocking function that checks for jobs every 'check_interval' seconds'''
		self.restore_all_job_logs()
		self._running_auto = True
		try:
			while self._running_auto:
				try:
					self.check()
					time.sleep(self._check_interval)
				except KeyboardInterrupt:
					print("KeyboardInterrupt")
					self.stop()
		finally:
			print("Stopping. Please wait, checking active async jobs ..")
			self.join()
		print(self, "Done!")

	def join(self):
		'''wait for any async jobs to complete'''
		for j in self.jobs:
			if isinstance(j, AsyncJobWrapper) and j.is_running: # wait for any running parallel tasks
				j.proc.join()
				print(j, "exited")

	def stop(self):
		'''stop job started with .start() method'''
		self._running_auto = False

	def get_job_by_id(self, jobid):
		for j in self.jobs:
			if j.jobid==jobid:
				return j
		return None

	def rerun(self, jobid, kwargs: dict=None):
		selected_job = self.get_job_by_id(jobid)
		if selected_job is None:
			raise IndexError("Invalid job id")
		if selected_job.is_running:
			raise RuntimeError("Cannot rerun a running task")
		if not isinstance(selected_job, AsyncJobWrapper):
			selected_job = AsyncJobWrapper(selected_job)
		selected_job.run(is_rerun=True, kwargs=kwargs)

	def disable_all(self):
		for j in self.jobs:
			j.disable()

	def enable_all(self):
		for j in self.jobs:
			j.enable()



import os
ROOT = os.path.dirname(os.path.realpath(__file__))
WEB_FOLDER = os.path.join(ROOT, 'web')


def _readTemplate(fileName, **kwargs):
	with open(fileName, 'r') as f:
		templateText = f.read()
	for arg in kwargs:
		templateText = templateText.replace("{{ "+arg+" }}", str(kwargs[arg]))
	return templateText


def HTML(content, title, css=[]):
	if not isinstance(css, (list,set,tuple)):
		css = [css]
	return _readTemplate(os.path.join(WEB_FOLDER, 'index.html'), title=title, body=str(content), body_css=' '.join(css))

def _TAG(tag, content, css, attrs):
	attrs = ['''{}="{}"'''.format(k,v) for k,v in attrs.items()]
	if not isinstance(css, (list,set,tuple)):
		css = [css]
	attrs.append('''class="{}"'''.format(' '.join(css)))
	return "<{t} {a}>{c}</{t}>".format(t=tag, a=' '.join(attrs), c=content)

def H(index, content, css=[], attrs={}):
	return _TAG('h{i}'.format(i=index), content, css, attrs)

def B(content, css=[], attrs={}):
	return _TAG('b', content, css, attrs)

def I(content, css=[], attrs={}):
	return _TAG('i', content, css, attrs)

def SMALL(content, css=[], attrs={}):
	return _TAG('small', content, css, attrs)

def SPAN(content, css=[], attrs={}):
	return _TAG('span', content, css, attrs)

def DIV(content, css=[], attrs={}):
	return _TAG('div', content, css, attrs)

def TABLE(thead='', tbody='', elem_id='', css=[], attrs={}):
	attrs['id'] = elem_id
	return _TAG('table', '{}{}'.format(thead, tbody), css, attrs)

def TH(h, default_sort=False):
	return "<th {}>{}</th>".format("data-sort-default" if default_sort else "", h)

def THEAD(th, css=[], attrs={}):
	return _TAG('thead', ''.join(th), css, attrs)

def TBODY(rows, css=[], attrs={}):
	return _TAG('tbody', ''.join(rows), css, attrs)

def TD(content, colspan=1, rowspan=1, css=[], attrs={}):
	attrs['colspan'] = colspan
	attrs['rowspan'] = rowspan
	return _TAG('td', content if content is not None else "-", css, attrs)

def TR(row, css=[], attrs={}):
	return _TAG('tr', ''.join(row), css, attrs)

def INPUT(content, css=[], attrs={}):
	return _TAG('input', content, css, attrs)

def CODE(s, css=[]):
	if not isinstance(css, (list,set,tuple)):
		css = [css]
	return '''<pre><code class="{}">{}</code></pre>'''.format(' '.join(css), s)

def SCRIPT(s):
	return "<script>{}</script>".format(s)


def SCRIPT_SRC(url):
	return '<script src="{}"></script>'.format(url)

def STYLE_LINK(url):
	return '<link rel="stylesheet" href="{}">'.format(url)
